from abc import ABC, abstractmethod
from cuqi.distribution import Distribution
import numpy as np
    
class RestorationPrior(Distribution):
    """    
    This class defines an implicit distribution associated with a restoration operator
    (eg denoiser). They are several works relating restorations operators with
    priors, see 
        -Laumont et al. https://arxiv.org/abs/2103.04715
        -Hu et al. https://openreview.net/pdf?id=x7d1qXEn1e
    We cannot sample from this distribution, neither compute its logpdf except in
    some cases. It allows us to apply algorithms such as MYULA and PnPULA.
    
    Parameters
    ---------- 
    restorator : callable f(x, restoration_strength)
        Function f that accepts input x to be restored and returns a two-element 
        tuple of the restored version of x and extra information about the 
        restoration operation. The second element can be of any type, including
        `None` in case there is no information.
            
    restorator_kwargs : dictionary
        Dictionary containing information about the restorator.
        It contains keyword argument parameters that will be passed to the
        restorator f. An example could be algorithm parameters such as the number
        of iterations or the stopping criteria.
    
    potential : callable function, optional
        The potential corresponds to the negative logpdf when it is accessible.
        This function is a mapping from the parameter domain to the real set. 
        It can be provided if the user knows how to relate it to the restorator.
        Ex: restorator is the proximal operator of the total variation (TV), then
        potential is the TV function. 
    """
    def __init__(self, restorator, restorator_kwargs
                =None, potential=None, **kwargs):
        if restorator_kwargs is None:
            restorator_kwargs = {}
        self.restorator = restorator 
        self.restorator_kwargs = restorator_kwargs
        self.potential = potential
        super().__init__(**kwargs)

    def restore(self, x, restoration_strength):
        """This function allows us to restore the input x with the user-supplied
        restorator. Extra information about the restoration operation is stored
        in the `RestorationPrior` info attribute.
        
        Parameters
        ---------- 
        x : ndarray
            parameter we want to restore.
        
        restoration_strength: positive float
            Strength of the restoration operation. In the case where the
            restorator is a denoiser, this parameter might correspond to the
            noise level.
        """
        restorator_return = self.restorator(x, restoration_strength=restoration_strength,
                                         **self.restorator_kwargs)

        if type(restorator_return) == tuple and len(restorator_return) == 2:
            solution, self.info = restorator_return
        else:
            raise ValueError("Unsupported return type from the user-supplied restorator function. "+ 
                             "Please ensure that the restorator function returns a two-element tuple with the "+
                             "restored solution as the first element and additional information about the "+
                             "restoration as the second element. The second element can be of any type, "+
                             "including `None` in case there is no particular information.")

        return solution
    
    def logpdf(self, x):
        """The logpdf function. It returns nan because we don't know the 
        logpdf of the implicit prior."""
        if self.potential is None:
            return np.nan
        else:
            return -self.potential(x)
    
    def _sample(self, N, rng=None):
        raise NotImplementedError("The sample method is not implemented for the"
                                  + "RestorationPrior class.")

    @property
    def _mutable_vars(self):
        """ Returns the mutable variables of the distribution. """
        # Currently mutable variables are not supported for user-defined
        # distributions.
        return []

    def get_conditioning_variables(self):
        """ Returns the conditioning variables of the distribution. """
        # Currently conditioning variables are not supported for user-defined
        # distributions.
        return []


class MoreauYoshidaPrior(Distribution):
    """    
    This class defines (implicit) smoothed priors for which we can apply 
    gradient-based algorithms. The smoothing is performed using
    the Moreau-Yoshida envelope of the target prior potential.
    
    In the following we give a detailed explanation of the
    Moreau-Yoshida smoothing.
    
    We consider a density such that - \log\pi(x) = -g(x) with g convex, lsc,
    proper but not differentiable. Consequently, we cannot apply any
    algorithm requiring the gradient of g.
    Idea:
    We consider the Moreau envelope of g defined as
    
    g_{smoothing_strength} (x) = inf_z 0.5*\| x-z \|_2^2/smoothing_strength + g(z).
    
    g_{smoothing_strength} has some nice properties
        - g_{smoothing_strength}(x)-->g(x) as smoothing_strength-->0 for all x
        - \nabla g_{smoothing_strength} is 1/smoothing_strength-Lipschitz
        - \nabla g_{smoothing_strength}(x) = (x - prox_g^{smoothing_strength}(x))/smoothing_strength for all x with 
        
        prox_g^{smoothing_strength}(x) = argmin_z 0.5*\| x-z \|_2^2/smoothing_strength + g(z) .

    Consequently, we can apply any gradient-based algorithm with
    g_{smoothing_strength} in lieu of g. These algorithms do not require the
    full knowledge of g_{smoothing_strength} but only its gradient. The gradient
    of g_{smoothing_strength} is fully determined by prox_g^{smoothing_strength}
    and smoothing_strength.
    It is important as, although there exists an explicit formula for
    g_{smoothing_strength}, it is rarely used in practice, as it would require
    us to solve an optimization problem each time we want to 
    estimate g_{smoothing_strength}. Furthermore, there exist cases where we dont't
    the regularization g with which the mapping prox_g^{smoothing_strength} is
    associated.

    Remark (Proximal operators are denoisers):
        We consider the denoising inverse problem x = u + n, with
        n \sim \mathcal{N}(0, smoothing_strength I).
        A mapping solving a denoising inverse problem is called denoiser. It takes
        the noisy observation x as an input and returns a less noisy version of x
        which is an estimate of u.
        We assume a prior density \pi(u) \propto exp(- g(u)).
        Then the MAP estimate is given by 
            x_MAP = \argmin_z 0.5 \| x - z \|_2^2/smoothing_strength + g(z) = prox_g^smoothing_strength(x)
        Then proximal operators are denoisers. 
    
    Remark (Denoisers are not necessarily proximal operators): Data-driven
    denoisers are not necessarily proximal operators
    (see https://arxiv.org/pdf/2201.13256)
    
    Parameters
    ---------- 
    prior : RestorationPrior
        Prior of the RestorationPrior type. In order to stay within the MYULA
        framework the restorator of RestorationPrior must be a proximal operator.
        
    smoothing_strength : float
        Smoothing strength of the Moreau-Yoshida envelope of the prior potential.
    """

    def __init__(self, prior:RestorationPrior, smoothing_strength=0.1, 
                 **kwargs):
        self.prior = prior
        self.smoothing_strength = smoothing_strength

        # if kwargs does not contain the geometry,
        # we set it to the geometry of the prior, if it exists
        if "geometry" in kwargs:
            raise ValueError(
                "The geometry parameter is not supported for the"
                + "MoreauYoshidaPrior class. The geometry is"
                + "automatically set to the geometry of the prior.")
        try:
            geometry = prior.geometry
        except:
            geometry = None

        super().__init__(geometry=geometry, **kwargs)

    @property
    def geometry(self):
        return self.prior.geometry

    @geometry.setter
    def geometry(self, value):
        self.prior.geometry = value

    @property
    def smoothing_strength(self):
        """ smoothing_strength of the distribution"""
        return self._smoothing_strength

    @smoothing_strength.setter
    def smoothing_strength(self, value):
        self._smoothing_strength = value
        
    @property
    def prior(self):
        """Getter for the MoreauYoshida prior."""
        return self._prior

    @prior.setter
    def prior(self, value):
        self._prior = value
    
    def gradient(self, x):
        """This is the gradient of the regularizer ie gradient of the negative
        logpdf of the implicit prior."""
        return -(x - self.prior.restore(x, self.smoothing_strength))/self.smoothing_strength
        
    def logpdf(self, x):
        """The logpdf function. It returns nan because we don't know the 
        logpdf of the implicit prior."""
        if self.prior.potential == None:
            return np.nan
        else:
            return -(self.prior.potential(self.prior.restore(x, self.smoothing_strength))*self.smoothing_strength +
                     0.5*((x-self.prior.restore(x, self.smoothing_strength))**2).sum())
    
    def _sample(self, N, rng=None):
        raise NotImplementedError("The sample method is not implemented for the"
                                  + f"{self.__class__.__name__} class.")

    @property
    def _mutable_vars(self):
        """ Returns the mutable variables of the distribution. """
        # Currently mutable variables are not supported for user-defined
        # distributions.
        return []

    def get_conditioning_variables(self):
        """ Returns the conditioning variables of the distribution. """
        # Currently conditioning variables are not supported for user-defined
        # distributions.
        return []

class TweediePrior(MoreauYoshidaPrior):
    """
    Alias for MoreauYoshidaPrior following Tweedie's formula framework. TweediePrior 
    defines priors where gradients are computed based on Tweedie's identity that links 
    MMSE (Minimum Mean Square Error) denoisers with the underlying smoothed prior, see:
        - Laumont et al. https://arxiv.org/abs/2103.04715 or https://doi.org/10.1137/21M1406349

    Tweedie's Formula
    -------------------------
    In the context of denoising, Tweedie's identity states that for a signal x
    corrupted by Gaussian noise:

        ∇_x log p_e(x) = (D_e(x) - x) / e

    where D_e(x) is the MMSE denoiser output and e is the noise variance.
    This enables us to perform gradient-based sampling with algorithms like ULA.

    At implementation level, TweediePrior shares identical functionality with MoreauYoshidaPrior. 
    Thus, it is implemented as an alias of MoreauYoshidaPrior, meaning all methods, 
    properties, and behavior are identical. The separate name provides clarity when 
    working specifically with Tweedie's formula-based approaches.

    Parameters
    ----------
    prior : RestorationPrior
        Prior of the RestorationPrior type containing a denoiser/restorator.
        
    smoothing_strength : float, default=0.1
        Corresponds to the noise variance e in Tweedie's formula context.

    See MoreauYoshidaPrior for the underlying implementation with complete documentation.
    """
    pass



"""
CUQIpy specific implementation of an abstract syntax tree (AST) for algebra on variables.

The AST is used to record the operations applied to variables allowing a delayed evaluation
of said operations when needed by traversing the tree with the __call__ method.

For example, the following code

    x = VariableNode('x')
    y = VariableNode('y')
    z = 2*x + 3*y

will create the following AST:

z = AddNode(
        MultiplyNode(
            ValueNode(2),
            VariableNode('x')
        ),
        MultiplyNode(
            ValueNode(3),
            VariableNode('y')
        )
    )

which can be evaluated by calling the __call__ method:

    z(x=1, y=2) # returns 8

"""

from abc import ABC, abstractmethod

convert_to_node = lambda x: x if isinstance(x, Node) else ValueNode(x)
""" Converts any non-Node object to a ValueNode object. """

# ====== Base classes for the nodes ======


class Node(ABC):
    """Base class for all nodes in the abstract syntax tree.

    Responsible for building the AST by creating nodes that represent the operations applied to variables.

    Each subclass must implement the __call__ method that will evaluate the node given the input parameters.

    """

    @abstractmethod
    def __call__(self, **kwargs):
        """Evaluate node at a given parameter value. This will traverse the sub-tree originated at this node and evaluate it given the recorded operations."""
        pass

    @abstractmethod
    def condition(self, **kwargs):
        """ Conditions the tree by replacing any VariableNode with a ValueNode if the variable is in the kwargs dictionary. """
        pass

    @abstractmethod
    def __repr__(self):
        """String representation of the node. Used for printing the AST."""
        pass

    def get_variables(self, variables=None):
        """Returns a set with the names of all variables in the sub-tree originated at this node."""
        if variables is None:
            variables = set()
        if isinstance(self, VariableNode):
            variables.add(self.name)
        if hasattr(self, "child"):
            self.child.get_variables(variables)
        if hasattr(self, "left"):
            self.left.get_variables(variables)
        if hasattr(self, "right"):
            self.right.get_variables(variables)
        return variables

    def __add__(self, other):
        return AddNode(self, convert_to_node(other))

    def __radd__(self, other):
        return AddNode(convert_to_node(other), self)

    def __sub__(self, other):
        return SubtractNode(self, convert_to_node(other))

    def __rsub__(self, other):
        return SubtractNode(convert_to_node(other), self)

    def __mul__(self, other):
        return MultiplyNode(self, convert_to_node(other))

    def __rmul__(self, other):
        return MultiplyNode(convert_to_node(other), self)

    def __truediv__(self, other):
        return DivideNode(self, convert_to_node(other))

    def __rtruediv__(self, other):
        return DivideNode(convert_to_node(other), self)

    def __pow__(self, other):
        return PowerNode(self, convert_to_node(other))

    def __rpow__(self, other):
        return PowerNode(convert_to_node(other), self)

    def __neg__(self):
        return NegateNode(self)

    def __abs__(self):
        return AbsNode(self)

    def __getitem__(self, i):
        return GetItemNode(self, convert_to_node(i))

    def __matmul__(self, other):
        return MatMulNode(self, convert_to_node(other))

    def __rmatmul__(self, other):
        return MatMulNode(convert_to_node(other), self)


class UnaryNode(Node, ABC):
    """Base class for all unary nodes in the abstract syntax tree.

    Parameters
    ----------
    child : Node
        The direct child node on which the unary operation is performed.

    """

    def __init__(self, child: Node):
        self.child = child

    def condition(self, **kwargs):
        return self.__class__(self.child.condition(**kwargs))


class BinaryNode(Node, ABC):
    """Base class for all binary nodes in the abstract syntax tree.

    The op_symbol attribute is used for printing the operation in the __repr__ method.

    Parameters
    ----------
    left : Node
        Left child node to the binary operation.

    right : Node
        Right child node to the binary operation.

    """

    @property
    @abstractmethod
    def op_symbol(self):
        """Symbol used to represent the operation in the __repr__ method."""
        pass

    def __init__(self, left: Node, right: Node):
        self.left = left
        self.right = right

    def condition(self, **kwargs):
        return self.__class__(self.left.condition(**kwargs), self.right.condition(**kwargs))

    def __repr__(self):
        return f"{self.left} {self.op_symbol} {self.right}"


class BinaryNodeWithParenthesis(BinaryNode, ABC):
    """Base class for all binary nodes in the abstract syntax tree that should be printed with parenthesis."""

    def __repr__(self):
        left = f"({self.left})" if isinstance(self.left, BinaryNode) else str(self.left)
        right = (
            f"({self.right})" if isinstance(self.right, BinaryNode) else str(self.right)
        )
        return f"{left} {self.op_symbol} {right}"

class BinaryNodeWithParenthesisNoSpace(BinaryNode, ABC):
    """Base class for all binary nodes in the abstract syntax tree that should be printed with parenthesis but no space."""

    def __repr__(self):
        left = f"({self.left})" if isinstance(self.left, BinaryNode) else str(self.left)
        right = (
            f"({self.right})" if isinstance(self.right, BinaryNode) else str(self.right)
        )
        return f"{left}{self.op_symbol}{right}"


# ====== Specific implementations of the "leaf" nodes ======


class VariableNode(Node):
    """Node that represents a generic variable, e.g. "x" or "y".

    Parameters
    ----------
    name : str
        Name of the variable. Used for printing and to retrieve the given input value
        of the variable in the kwargs dictionary when evaluating the tree.

    """

    def __init__(self, name):
        self.name = name

    def __call__(self, **kwargs):
        """Retrieves the value of the variable from the passed kwargs. If no value is found, it raises a KeyError."""
        if not self.name in kwargs:
            raise KeyError(
                f"Variable '{self.name}' not found in the given input parameters. Unable to evaluate the expression."
            )
        return kwargs[self.name]

    def condition(self, **kwargs):
        if self.name in kwargs:
            return ValueNode(kwargs[self.name])
        return self

    def __repr__(self):
        return self.name


class ValueNode(Node):
    """Node that represents a constant value. The value can be any python object that is not a Node.

    Parameters
    ----------
    value : object
        The python object that represents the value of the node.

    """

    def __init__(self, value):
        self.value = value

    def __call__(self, **kwargs):
        """Returns the value of the node."""
        return self.value

    def condition(self, **kwargs):
        return self

    def __repr__(self):
        return str(self.value)


# ====== Specific implementations of the "internal" nodes ======


class AddNode(BinaryNode):
    """Node that represents the addition operation."""

    @property
    def op_symbol(self):
        return "+"

    def __call__(self, **kwargs):
        return self.left(**kwargs) + self.right(**kwargs)


class SubtractNode(BinaryNode):
    """Node that represents the subtraction operation."""

    @property
    def op_symbol(self):
        return "-"

    def __call__(self, **kwargs):
        return self.left(**kwargs) - self.right(**kwargs)


class MultiplyNode(BinaryNodeWithParenthesis):
    """Node that represents the multiplication operation."""

    @property
    def op_symbol(self):
        return "*"

    def __call__(self, **kwargs):
        return self.left(**kwargs) * self.right(**kwargs)


class DivideNode(BinaryNodeWithParenthesis):
    """Node that represents the division operation."""

    @property
    def op_symbol(self):
        return "/"

    def __call__(self, **kwargs):
        return self.left(**kwargs) / self.right(**kwargs)


class PowerNode(BinaryNodeWithParenthesisNoSpace):
    """Node that represents the power operation."""

    @property
    def op_symbol(self):
        return "^"

    def __call__(self, **kwargs):
        return self.left(**kwargs) ** self.right(**kwargs)


class GetItemNode(BinaryNode):
    """Node that represents the get item operation. Here the left node is the object and the right node is the index."""

    def __call__(self, **kwargs):
        return self.left(**kwargs)[self.right(**kwargs)]

    def __repr__(self):
        left = f"({self.left})" if isinstance(self.left, BinaryNode) else str(self.left)
        return f"{left}[{self.right}]"
    
    @property
    def op_symbol(self):
        pass


class NegateNode(UnaryNode):
    """Node that represents the arithmetic negation operation."""

    def __call__(self, **kwargs):
        return -self.child(**kwargs)

    def __repr__(self):
        child = (
            f"({self.child})"
            if isinstance(self.child, (BinaryNode, UnaryNode))
            else str(self.child)
        )
        return f"-{child}"


class AbsNode(UnaryNode):
    """Node that represents the absolute value operation."""

    def __call__(self, **kwargs):
        return abs(self.child(**kwargs))

    def __repr__(self):
        return f"abs({self.child})"


class MatMulNode(BinaryNodeWithParenthesis):
    """Node that represents the matrix multiplication operation."""

    @property
    def op_symbol(self):
        return "@"

    def __call__(self, **kwargs):
        return self.left(**kwargs) @ self.right(**kwargs)



from __future__ import annotations
import numpy as np
from scipy.sparse import csc_matrix
from scipy.sparse import hstack
from scipy.linalg import solve
from cuqi.samples import Samples
from cuqi.array import CUQIarray
from cuqi.geometry import Geometry, _DefaultGeometry1D, _DefaultGeometry2D,\
    _get_identity_geometries
import cuqi
import matplotlib.pyplot as plt
from copy import copy
from functools import partial
from cuqi.utilities import force_ndarray

class Model(object):
    """Generic model defined by a forward operator.

    Parameters
    -----------
    forward : callable function
        Forward operator of the model. It takes one or more inputs and returns the model output.

    range_geometry : integer, a 1D or 2D tuple of integers, cuqi.geometry.Geometry 
        If integer or 1D tuple of integers is given, a cuqi.geometry._DefaultGeometry1D is created with dimension of the integer.
        If 2D tuple of integers is given, a cuqi.geometry._DefaultGeometry2D is created with dimensions of the tuple.
        If cuqi.geometry.Geometry object is given, it is used as the range geometry of the model.

    domain_geometry : integer, a 1D or 2D tuple of integers, cuqi.geometry.Geometry or a tuple with items of any of the listed types
        If integer or 1D tuple of integers is given, a cuqi.geometry._DefaultGeometry1D is created with dimension of the integer.
        If 2D tuple of integers is given (and the forward model has one input only), a cuqi.geometry._DefaultGeometry2D is created with dimensions of the tuple.
        If cuqi.geometry.Geometry is given, it is used as the domain geometry.
        If tuple of the above types is given, a cuqi.geometry._ProductGeometry is created based on the tuple entries. This is used for models with multiple inputs where each entry in the tuple represents the geometry of each input.

    gradient : callable function, a tuple of callable functions or None, optional
        The direction-Jacobian product of the forward model Jacobian with respect to the model input, evaluated at the model input. For example, if the forward model inputs are `x` and `y`, the gradient callable signature should be (`direction`, `x`, `y`), in that order, where `direction` is the direction by which the Jacobian matrix is multiplied and `x` and `y` are the parameters at which the Jacobian is computed.

        If the gradient function is a single callable function, it returns a 1D ndarray if the model has only one input. If the model has multiple inputs, this gradient function should return a tuple of 1D ndarrays, each representing the gradient with respect to each input.

        If the gradient function is a tuple of callable functions, each callable function should return a 1D ndarray representing the gradient with respect to each input. The order of the callable functions in the tuple should match the order of the model inputs.

    jacobian : callable function, a tuple of callable functions or None, optional
        The Jacobian of the forward model with respect to the forward model input, evaluated at the model input. For example, if the forward model inputs are `x` and `y`, the jacobian signature should be (`x`, `y`), in that order, where `x` and `y` are the parameters at which the Jacobian is computed.

        If the Jacobian function is a single callable function, it should return a 2D ndarray of shape (range_dim, domain_dim) if the model has only one input. If the model has multiple inputs, this Jacobian function should return a tuple of 2D ndarrays, each representing the Jacobian with respect to each input. 
        
        If the Jacobian function is a tuple of callable functions, each callable function should return a 2D ndarray representing the Jacobian with respect to each input. The order of the callable functions in the tuple should match the order of the model inputs.
    
        The Jacobian function is used to specify the gradient function by computing the vector-Jacobian product (VJP), here we refer to the vector in the VJP as the `direction` since it is the direction at which the gradient is computed. Either the gradient or the Jacobian can be specified, but not both.


    :ivar range_geometry: The geometry representing the range.
    :ivar domain_geometry: The geometry representing the domain.

    Example 1
    ----------

    Consider a forward model :math:`F: \mathbb{R}^2 \\rightarrow \mathbb{R}` defined by the following forward operator:

    .. math::

        F(x) = 10x_2 - 10x_1^3 + 5x_1^2 + 6x_1

    The jacobian matrix of the forward operator is given by:

    .. math::

        J_F(x) = \\begin{bmatrix} -30x_1^2 + 10x_1 + 6 & 10 \\end{bmatrix}

    The forward model can be defined as follows:

    .. code-block:: python

        import numpy as np
        from cuqi.model import Model

        def forward(x):
            return 10*x[1] - 10*x[0]**3 + 5*x[0]**2 + 6*x[0]

        def jacobian(x): # Can use "x" or "wrt" as the input argument name
            return np.array([[-30*x[0]**2 + 10*x[0] + 6, 10]])

        model = Model(forward, range_geometry=1, domain_geometry=2, jacobian=jacobian)

        print(model(np.array([1, 1])))
        print(model.gradient(np.array([1]), np.array([1, 1])))

    Alternatively, the gradient information in the forward model can be defined by direction-Jacobian product using the gradient keyword argument.

    This may be more efficient if forming the Jacobian matrix is expensive.

    .. code-block:: python

        import numpy as np
        from cuqi.model import Model

        def forward(x):
            return 10*x[1] - 10*x[0]**3 + 5*x[0]**2 + 6*x[0]

        def gradient(direction, x):
            # Direction-Jacobian product direction@jacobian(x)
            return direction@np.array([[-30*x[0]**2 + 10*x[0] + 6, 10]])

        model = Model(forward, range_geometry=1, domain_geometry=2, gradient=gradient)

        print(model(np.array([1, 1])))
        print(model.gradient(np.array([1]), np.array([1, 1])))

    Example 2
    ----------
    Alternatively, the example above can be defined as a model with multiple inputs: :math:`x` and :math:`y`:

    .. code-block:: python

        import numpy as np
        from cuqi.model import Model
        from cuqi.geometry import Discrete
    
        def forward(x, y):
            return 10 * y - 10 * x**3 + 5 * x**2 + 6 * x
    
        def jacobian(x, y):
            return (np.array([[-30 * x**2 + 10 * x + 6]]), np.array([[10]]))
    
        model = Model(
            forward,
            range_geometry=1,
            domain_geometry=(Discrete(1), Discrete(1)),
            jacobian=jacobian,
        )
    
        print(model(1, 1))
        print(model.gradient(np.array([1]), 1, 1))
    """

    _supports_partial_eval = True
    """Flag indicating that partial evaluation of Model objects is supported, i.e., calling the model object with only some of the inputs specified returns a model that can be called with the remaining inputs."""

    def __init__(self, forward, range_geometry, domain_geometry, gradient=None, jacobian=None):

        # Check if input is callable
        if callable(forward) is not True:
            raise TypeError("Forward needs to be callable function.")

        # Store forward func
        self._forward_func = forward
        self._stored_non_default_args = None

        # Store range_geometry
        self.range_geometry = range_geometry

        # Store domain_geometry
        self.domain_geometry = domain_geometry

        # Additional checks for the forward operator
        self._check_domain_geometry_consistent_with_forward()

        # Check if only one of gradient and jacobian is given
        if (gradient is not None) and (jacobian is not None):
            raise TypeError("Only one of gradient and jacobian should be specified")

        # Check correct gradient form (check type, signature, etc.)
        self._check_correct_gradient_jacobian_form(gradient, "gradient")

        # Check correct jacobian form (check type, signature, etc.)
        self._check_correct_gradient_jacobian_form(jacobian, "jacobian")

        # If jacobian is provided, use it to specify gradient function
        # (vector-Jacobian product)
        if jacobian is not None:
            gradient = self._use_jacobian_to_specify_gradient(jacobian)

        self._gradient_func = gradient

        # Set gradient output stacked flag to False
        self._gradient_output_stacked = False

    @property
    def _non_default_args(self):
        if self._stored_non_default_args is None:
            # Store non_default_args of the forward operator for faster caching
            # when checking for those arguments.
            self._stored_non_default_args =\
                cuqi.utilities.get_non_default_args(self._forward_func)
        return self._stored_non_default_args

    @property
    def number_of_inputs(self):
        """ The number of inputs of the model. """
        return len(self._non_default_args)

    @property
    def range_geometry(self):
        """ The geometry representing the range of the model. """
        return self._range_geometry

    @range_geometry.setter
    def range_geometry(self, value):
        """ Update the range geometry of the model. """
        if isinstance(value, Geometry):
            self._range_geometry = value
        elif isinstance(value, int):
            self._range_geometry = self._create_default_geometry(value)
        elif isinstance(value, tuple):
            self._range_geometry = self._create_default_geometry(value)
        elif value is None:
            raise AttributeError(
                "The parameter 'range_geometry' is not specified by the user and it cannot be inferred from the attribute 'forward'."
            )
        else:
            raise TypeError(
                " The allowed types for 'range_geometry' are: 'cuqi.geometry.Geometry', int, 1D tuple of int, or 2D tuple of int."
            )

    @property
    def domain_geometry(self):
        """ The geometry representing the domain of the model. """
        return self._domain_geometry

    @domain_geometry.setter
    def domain_geometry(self, value):
        """ Update the domain geometry of the model. """

        if isinstance(value, Geometry):
            self._domain_geometry = value
        elif isinstance(value, int):
            self._domain_geometry = self._create_default_geometry(value)
        elif isinstance(value, tuple) and self.number_of_inputs == 1:
            self._domain_geometry = self._create_default_geometry(value)
        elif isinstance(value, tuple) and self.number_of_inputs > 1:
            geometries = [item if isinstance(item, Geometry) else self._create_default_geometry(item) for item in value]
            self._domain_geometry = cuqi.experimental.geometry._ProductGeometry(*geometries)
        elif value is None:
            raise AttributeError(
                "The parameter 'domain_geometry' is not specified by the user and it cannot be inferred from the attribute 'forward'."
            )
        else:
            raise TypeError(
                "For forward model with 1 input, the allowed types for 'domain_geometry' are: 'cuqi.geometry.Geometry', int, 1D tuple of int, or 2D tuple of int. For forward model with multiple inputs, the 'domain_geometry' should be a tuple with items of any of the above types."
            )

    def _create_default_geometry(self, value):
        """Private function that creates default geometries for the model."""
        if isinstance(value, tuple) and len(value) == 1:
            value = value[0]
        if isinstance(value, Geometry):
            return value
        if isinstance(value, int):
            return _DefaultGeometry1D(grid=value)
        elif isinstance(value, tuple) and len(value) == 2:
            return _DefaultGeometry2D(im_shape=value)
        else:
            raise ValueError(
                "Default geometry creation can be specified by an integer or a 2D tuple of integers."
            )

    @property
    def domain_dim(self):
        """
        The dimension of the domain
        """
        return self.domain_geometry.par_dim

    @property
    def range_dim(self): 
        """
        The dimension of the range
        """
        return self.range_geometry.par_dim

    def _check_domain_geometry_consistent_with_forward(self):
        """Private function that checks if the domain geometry of the model is
        consistent with the forward operator."""
        if (
            not isinstance(
                self.domain_geometry, cuqi.experimental.geometry._ProductGeometry
            )
            and self.number_of_inputs > 1
        ):
            raise ValueError(
                "The forward operator input is specified by more than one argument. This is only supported for domain geometry of type tuple with items of type: cuqi.geometry.Geometry object, int, or 2D tuple of int."
            )

    def _check_correct_gradient_jacobian_form(self, func, func_type):
        """Private function that checks if the gradient/jacobian parameter is
        in the correct form. That is, check if the gradient/jacobian has the
        correct type, signature, etc."""

        if func is None:
            return

        # gradient/jacobian should be callable (for single input and multiple input case)
        # or a tuple of callables (for multiple inputs case)
        if isinstance(func, tuple):
            # tuple length should be same as the number of inputs
            if len(func) != self.number_of_inputs:
                raise ValueError(
                    f"The "
                    + func_type.lower()
                    + f" tuple length should be {self.number_of_inputs} for model with inputs {self._non_default_args}"
                )
            # tuple items should be callables or None
            if not all([callable(func_i) or func_i is None for func_i in func]):
                raise TypeError(
                    func_type.capitalize()
                    + " tuple should contain callable functions or None."
                )

        elif callable(func):
            # temporarily convert gradient/jacobian to tuple for checking only
            func = (func,)

        else:
            raise TypeError(
                "Gradient needs to be callable function or tuple of callable functions."
            )

        expected_func_non_default_args = (
            self._non_default_args
            if not hasattr(self, "_original_non_default_args")
            else self._original_non_default_args
        )

        if func_type.lower() == "gradient":
            # prepend 'direction' to the expected gradient non default args
            expected_func_non_default_args = [
                "direction"
            ] + expected_func_non_default_args

        for func_i in func:
            # make sure the signature of the gradient/jacobian function is correct
            # that is, the same as the expected_func_non_default_args
            if func_i is not None:
                func_non_default_args = cuqi.utilities.get_non_default_args(func_i)

                if list(func_non_default_args) != list(expected_func_non_default_args):
                    raise ValueError(
                        func_type.capitalize()
                        + f" function signature should be {expected_func_non_default_args}"
                    )

    def _use_jacobian_to_specify_gradient(self, jacobian):
        """Private function that uses the jacobian function to specify the
        gradient function."""
        # if jacobian is a single function and model has multiple inputs
        if callable(jacobian) and self.number_of_inputs > 1:
            gradient = self._create_gradient_lambda_function_from_jacobian_with_correct_signature(
                jacobian, form='one_callable_multiple_inputs'
            )
        # Elif jacobian is a single function and model has only one input
        elif callable(jacobian):
            gradient = self._create_gradient_lambda_function_from_jacobian_with_correct_signature(
                jacobian, form='one_callable_one_input'
            )
        # Else, jacobian is a tuple of jacobian functions
        else:
            gradient = []
            for jac in jacobian:
                if jac is not None:
                    gradient.append(
                        self._create_gradient_lambda_function_from_jacobian_with_correct_signature(
                            jac, form='tuple_of_callables'
                        )
                    )
                else:
                    gradient.append(None)
        return tuple(gradient) if isinstance(gradient, list) else gradient

    def _create_gradient_lambda_function_from_jacobian_with_correct_signature(
        self, jacobian, form
    ):
        """Private function that creates gradient lambda function from the
        jacobian function, with the correct signature (based on the model
        non_default_args).
        """
        # create the string representation of the lambda function
        # for different forms of jacobian
        if form=='one_callable_multiple_inputs':
            grad_fun_str = (
                "lambda direction, "
                + ", ".join(self._non_default_args)
                + ", jacobian: tuple([direction@jacobian("
                + ", ".join(self._non_default_args)
                + ")[i] for i in range("+str(self.number_of_inputs)+")])"
            )
        elif form=='tuple_of_callables' or form=='one_callable_one_input':
            grad_fun_str = (
                "lambda direction, "
                + ", ".join(self._non_default_args)
                + ", jacobian: direction@jacobian("
                + ", ".join(self._non_default_args)
                + ")"
            )
        else:
            raise ValueError("form should be either 'one_callable' or 'tuple_of_callables'.")

        # create the lambda function from the string
        grad_func = eval(grad_fun_str)

        # create partial function from the lambda function with jacobian as a
        # fixed argument
        grad_func = partial(grad_func, jacobian=jacobian)

        return grad_func

    def _2fun(self, geometry=None, is_par=True, **kwargs):
        """ Converts `kwargs` to function values (if needed) using the geometry. For example, `kwargs` can be the model input which need to be converted to function value before being passed to :class:`~cuqi.model.Model` operators (e.g. _forward_func, _adjoint_func, _gradient_func).

        Parameters
        ----------
        geometry : cuqi.geometry.Geometry
            The geometry representing the values in `kwargs`.

        is_par : bool or a tuple of bools
            If `is_par` is True, the values in `kwargs` are assumed to be parameters.
            If `is_par` is False, the values in `kwargs` are assumed to be function values.
            If `is_par` is a tuple of bools, the values in `kwargs` are assumed to be parameters or function values based on the corresponding boolean value in the tuple.
        
        **kwargs : keyword arguments to be converted to function values.

        Returns
        -------
        dict of the converted values
        """
        # Check kwargs and geometry are consistent and set up geometries list and
        # is_par tuple
        geometries, is_par = self._helper_pre_conversion_checks_and_processing(geometry, is_par, **kwargs)

        # Convert to function values
        for i, (k, v) in enumerate(kwargs.items()):
            # Use CUQIarray funvals if geometry is consistent
            if isinstance(v, CUQIarray) and v.geometry == geometries[i]:
                kwargs[k] = v.funvals
            # Else, if we still need to convert to function value (is_par[i] is True)
            # we use the geometry par2fun method
            elif is_par[i] and v is not None:
                kwargs[k] = geometries[i].par2fun(v)
            else:
                # No need to convert
                pass

        return kwargs

    def _helper_pre_conversion_checks_and_processing(self, geometry=None, is_par=True, **kwargs):
        """ Helper function that checks if kwargs and geometry are consistent
        and sets up geometries list and is_par tuple.
        """
        # If len of kwargs is larger than 1, the geometry needs to be of type
        # _ProductGeometry
        if (
            not isinstance(geometry, cuqi.experimental.geometry._ProductGeometry)
            and len(kwargs) > 1
        ):
            raise ValueError(
                "The input is specified by more than one argument. This is only "
                + "supported for domain geometry of type "
                + f"{cuqi.experimental.geometry._ProductGeometry.__name__}."
            )

        # If is_par is bool, make it a tuple of bools of the same length as
        # kwargs
        is_par = (is_par,) * len(kwargs) if isinstance(is_par, bool) else is_par

        # Set up geometries list
        geometries = (
            geometry.geometries
            if isinstance(geometry, cuqi.experimental.geometry._ProductGeometry)
            else [geometry]
        )

        return geometries, is_par

    def _2par(self, geometry=None, to_CUQIarray=False, is_par=False, **kwargs):    
        """ Converts `kwargs` to parameters using the geometry. For example, `kwargs` can be the output of :class:`~cuqi.model.Model` operators (e.g. _forward_func, _adjoint_func, _gradient_func) which need to be converted to parameters before being returned.

        Parameters
        ----------
        geometry : cuqi.geometry.Geometry
            The geometry representing the values in `kwargs`.

        to_CUQIarray : bool or a tuple of bools
            If `to_CUQIarray` is True, the values in `kwargs` will be wrapped in `CUQIarray`.
            If `to_CUQIarray` is False, the values in `kwargs` will not be wrapped in `CUQIarray`.
            If `to_CUQIarray` is a tuple of bools, the values in `kwargs` will be wrapped in `CUQIarray` or not based on the corresponding boolean value in the tuple.
        
        is_par : bool or a tuple of bools
            If `is_par` is True, the values in `kwargs` are assumed to be parameters.
            If `is_par` is False, the values in `kwargs` are assumed to be function values.
            If `is_par` is a tuple of bools, the values in `kwargs` are assumed to be parameters or function values based on the corresponding boolean value in the tuple.

        Returns
        -------
        dict of the converted values
        """
        # Check kwargs and geometry are consistent and set up geometries list and
        # is_par tuple
        geometries, is_par = self._helper_pre_conversion_checks_and_processing(geometry, is_par, **kwargs)

        # if to_CUQIarray is bool, make it a tuple of bools of the same length
        # as kwargs
        to_CUQIarray = (to_CUQIarray,) * len(kwargs) if isinstance(to_CUQIarray, bool) else to_CUQIarray

        # Convert to parameters
        for i , (k, v) in enumerate(kwargs.items()):
            # Use CUQIarray parameters if geometry is consistent
            if isinstance(v, CUQIarray) and v.geometry == geometries[i]:
                v = v.parameters
            # Else, if we still need to convert to parameter value (is_par[i] is False)
            # we use the geometry fun2par method
            elif not is_par[i] and v is not None:
                v = geometries[i].fun2par(v)
            else:
                # No need to convert
                pass

            # Wrap the value v in CUQIarray if requested
            if to_CUQIarray[i] and v is not None:
                v = CUQIarray(v, is_par=True, geometry=geometries[i])

            kwargs[k] = v

        return kwargs

    def _apply_func(self, func=None, fwd=True, is_par=True, **kwargs):
        """ Private function that applies the given function `func` to the input `kwargs`. It converts the input to function values (if needed) and converts the output to parameter values. It additionally handles the case of applying the function `func` to cuqi.samples.Samples objects.
        
        Parameters
        ----------
        func: function handler 
            The function to be applied.

        fwd : bool
            Flag indicating the direction of the operator to determine the range and domain geometries of the function.
            If True the function is a forward operator.
            If False the function is an adjoint operator.

        is_par : bool or list of bool
            If True, the inputs in `kwargs` are assumed to be parameters.
            If False, the input in `kwargs` are assumed to be function values.
            If `is_par` is a list of bools, the inputs are assumed to be parameters or function values based on the corresponding boolean value in the list.

        Returns
        -------
        ndarray or cuqi.array.CUQIarray or cuqi.samples.Samples object
            The output of the function.
        """ 
        # Specify the range and domain geometries of the function
        # If forward operator, range geometry is the model range geometry and
        # domain geometry is the model domain geometry
        if fwd:
            func_range_geometry = self.range_geometry
            func_domain_geometry = self.domain_geometry
        # If adjoint operator, range geometry is the model domain geometry and
        # domain geometry is the model range geometry
        else:
            func_range_geometry = self.domain_geometry
            func_domain_geometry = self.range_geometry

        # If input x is Samples we apply func for each sample
        # TODO: Check if this can be done all-at-once for computational speed-up
        if any(isinstance(x, Samples) for x in kwargs.values()):
            return self._handle_case_when_model_input_is_samples(func, fwd, **kwargs)

        # store if any input x is CUQIarray
        is_CUQIarray = any(isinstance(x, CUQIarray) for x in kwargs.values())

        # Convert input to function values
        kwargs = self._2fun(geometry=func_domain_geometry, is_par=is_par, **kwargs)

        # Apply the function
        out = func(**kwargs)

        # Return output as parameters
        # (wrapped in CUQIarray if any input was CUQIarray)
        return self._2par(
            geometry=func_range_geometry, to_CUQIarray=is_CUQIarray, **{"out": out}
        )["out"]

    def _handle_case_when_model_input_is_samples(self, func=None, fwd=True, **kwargs):
        """Private function that calls apply_func for samples in the
        Samples object(s).
        """
        # All kwargs should be Samples objects
        if not all(isinstance(x, Samples) for x in kwargs.values()):
            raise TypeError(
                "If applying the function to Samples, all inputs should be Samples."
            )

        # All Samples objects should have the same number of samples
        Ns = list(kwargs.values())[0].Ns
        if not all(x.Ns == Ns for x in kwargs.values()):
            raise ValueError(
                "If applying the function to Samples, all inputs should have the same number of samples."
            )

        # Specify the range dimension of the function
        range_dim = self.range_dim if fwd else self.domain_dim

        # Create empty array to store the output
        out = np.zeros((range_dim, Ns))

        # Recursively apply func to each sample
        for i in range(Ns):
            kwargs_i = {
                k: CUQIarray(v.samples[..., i], is_par=v.is_par, geometry=v.geometry)
                for k, v in kwargs.items()
            }
            out[:, i] = self._apply_func(func=func, fwd=fwd, **kwargs_i)
        # Specify the range geometries of the function
        func_range_geometry = self.range_geometry if fwd else self.domain_geometry
        return Samples(out, geometry=func_range_geometry)

    def _parse_args_add_to_kwargs(
        self, *args, is_par=True, non_default_args=None, map_name="model", **kwargs
    ):
        """ Private function that parses the input arguments and adds them as
        keyword arguments matching (the order of) the non default arguments of
        the forward function or other specified non_default_args list.
        """
        # If non_default_args is not specified, use the non_default_args of the
        # model
        if non_default_args is None:
            non_default_args = self._non_default_args

        # Either args or kwargs can be provided but not both
        if len(args) > 0 and len(kwargs) > 0:
            raise ValueError(
                "The "
                + map_name.lower()
                + " input is specified both as positional and keyword arguments. This is not supported."
            )

        len_input = len(args) + len(kwargs)

        # If partial evaluation, make sure input is not of type Samples
        if len_input < len(non_default_args):
            # If the argument is a Sample object, splitting or partial
            # evaluation of the model is not supported
            temp_args = args if len(args) > 0 else list(kwargs.values())
            if any(isinstance(arg, Samples) for arg in temp_args):
                raise ValueError(("When using Samples objects as input, the"
                                +" user should provide a Samples object for"
                                +f" each non_default_args {non_default_args}"
                                +" of the model. That is, partial evaluation"
                                +" or splitting is not supported for input"
                                +" of type Samples."))

        # If args are given, add them to kwargs
        if len(args) > 0:

            # Check if the input is for multiple input case and is stacked,
            # then split it
            if len(args) < len(non_default_args):
                args = self._split_in_case_of_stacked_args(*args, is_par=is_par)

            # Add args to kwargs following the order of non_default_args
            for idx, arg in enumerate(args):
                kwargs[non_default_args[idx]] = arg
    
        # Check kwargs matches non_default_args
        if not (set(list(kwargs.keys())) <= set(non_default_args)):
            if map_name == "gradient":
                error_msg = f"The gradient input is specified by a direction and keywords arguments {list(kwargs.keys())} that does not match the non_default_args of the model {non_default_args}."
            else:
                error_msg = (
                    "The "
                    + map_name.lower()
                    + f" input is specified by keywords arguments {list(kwargs.keys())} that does not match the non_default_args of the "
                    + map_name
                    + f" {non_default_args}."
                )

            raise ValueError(error_msg)

        # Make sure order of kwargs is the same as non_default_args
        kwargs = {k: kwargs[k] for k in non_default_args if k in kwargs}

        return kwargs

    def _split_in_case_of_stacked_args(self, *args, is_par=True):
        """Private function that checks if the input args is a stacked
        CUQIarray or numpy array and splits it into multiple arguments based on
        the domain geometry of the model. Otherwise, it returns the input args
        unchanged."""

        # Check conditions for splitting and split if all conditions are met
        is_CUQIarray = isinstance(args[0], CUQIarray)
        is_numpy_array = isinstance(args[0], np.ndarray)

        if ((is_CUQIarray or is_numpy_array) and
           is_par and
           len(args) == 1 and
           args[0].shape == (self.domain_dim,) and
           isinstance(self.domain_geometry, cuqi.experimental.geometry._ProductGeometry)):
            # Split the stacked input
            split_args = np.split(args[0], self.domain_geometry.stacked_par_split_indices)
            # Convert split args to CUQIarray if input is CUQIarray
            if is_CUQIarray:
                split_args = [
                    CUQIarray(arg, is_par=True, geometry=self.domain_geometry.geometries[i])
                    for i, arg in enumerate(split_args)
                ]
            return split_args

        else:
            return args

    def forward(self, *args, is_par=True, **kwargs):
        """ Forward function of the model.

        Forward converts the input to function values (if needed) using the domain geometry of the model. Then it applies the forward operator to the function values and converts the output to parameters using the range geometry of the model.

        Parameters
        ----------
        *args : ndarrays or cuqi.array.CUQIarray objects or cuqi.samples.Samples objects
            Positional arguments for the forward operator. The forward operator input can be specified as either positional arguments or keyword arguments but not both.

            If the input is specified as positional arguments, the order of the arguments should match the non_default_args of the model.

        is_par : bool or a tuple of bools
            If True, the inputs in `args` or `kwargs` are assumed to be parameters.
            If False, the inputs in `args` or `kwargs` are assumed to be function values.
            If `is_par` is a tuple of bools, the inputs are assumed to be parameters or function values based on the corresponding boolean value in the tuple.

        **kwargs : keyword arguments
            keyword arguments for the forward operator. The forward operator input can be specified as either positional arguments or keyword arguments but not both.

            If the input is specified as keyword arguments, the keys should match the non_default_args of the model.

        Returns
        -------
        ndarray or cuqi.array.CUQIarray or cuqi.samples.Samples object
            The model output. Always returned as parameters.
        """

        # Add args to kwargs and ensure the order of the arguments matches the
        # non_default_args of the forward function
        kwargs = self._parse_args_add_to_kwargs(
            *args, **kwargs, is_par=is_par, map_name="model"
        )
        # Extract args from kwargs
        args = list(kwargs.values())

        if len(kwargs) == 0:
            return self

        partial_arguments = len(kwargs) < len(self._non_default_args)

        # If input is a distribution, we simply change the parameter name of
        # model to match the distribution name
        if all(isinstance(x, cuqi.distribution.Distribution)
               for x in kwargs.values()):
            if partial_arguments:
                raise ValueError(
                    "Partial evaluation of the model is not supported for distributions."
                )
            return self._handle_case_when_model_input_is_distributions(kwargs)

        # If input is a random variable, we handle it separately
        elif all(isinstance(x, cuqi.experimental.algebra.RandomVariable)
               for x in kwargs.values()):
            if partial_arguments:
                raise ValueError(
                    "Partial evaluation of the model is not supported for random variables."
                )
            return self._handle_case_when_model_input_is_random_variables(kwargs)

        # If input is a Node from internal abstract syntax tree, we let the Node handle the operation
        # We use NotImplemented to indicate that the operation is not supported from the Model class
        # in case of operations such as "@" that can be interpreted as both __matmul__ and __rmatmul__
        # the operation may be delegated to the Node class.
        elif any(isinstance(args_i, cuqi.experimental.algebra.Node) for args_i in args):
            return NotImplemented

        # if input is partial, we create a new model with the partial input
        if partial_arguments:
            # Create is_par_partial from the is_par to contain only the relevant parts
            if isinstance(is_par, (list, tuple)):
                is_par_partial = tuple(
                    is_par[i]
                    for i in range(self.number_of_inputs)
                    if self._non_default_args[i] in kwargs.keys()
                )
            else:
                is_par_partial = is_par
            # Build a partial model with the given kwargs
            partial_model = self._build_partial_model(kwargs, is_par_partial)
            return partial_model

        # Else we apply the forward operator
        # if model has _original_non_default_args, we use it to replace the
        # kwargs keys so that it matches self._forward_func signature
        if hasattr(self, '_original_non_default_args'):
            kwargs = {k:v for k,v in zip(self._original_non_default_args, args)}
        return self._apply_func(func=self._forward_func,
                                fwd=True,
                                is_par=is_par,
                                **kwargs)

    def _correct_distribution_dimension(self, distributions):
        """Private function that checks if the dimension of the
        distributions matches the domain dimension of the model."""
        if len(distributions) == 1:
            return list(distributions)[0].dim == self.domain_dim
        elif len(distributions) > 1 and isinstance(
            self.domain_geometry, cuqi.experimental.geometry._ProductGeometry
        ):
            return all(
                d.dim == self.domain_geometry.par_dim_list[i]
                for i, d in enumerate(distributions)
            )
        else:
            return False

    def _build_partial_model(self, kwargs, is_par):
        """Private function that builds a partial model substituting the given
        keyword arguments with their values. The created partial model will have
        as inputs the non-default arguments that are not in the kwargs."""

        # Extract args from kwargs
        args = list(kwargs.values())

        # Define original_non_default_args which represents the complete list of
        # non-default arguments of the forward function.
        original_non_default_args = (
            self._original_non_default_args
            if hasattr(self, "_original_non_default_args")
            else self._non_default_args
        )

        if hasattr(self, "_original_non_default_args"):
            # Split the _original_non_default_args into two lists:
            # 1. reduced_original_non_default_args: the _original_non_default_args
            # corresponding to the _non_default_args that are not in kwargs
            # 2. substituted_non_default_args: the _original_non_default_args
            # corresponding to the _non_default_args that are in kwargs
            reduced_original_non_default_args = [
                original_non_default_args[i]
                for i in range(self.number_of_inputs)
                if self._non_default_args[i] not in kwargs.keys()
            ]
            substituted_non_default_args = [
                original_non_default_args[i]
                for i in range(self.number_of_inputs)
                if self._non_default_args[i] in kwargs.keys()
            ]
            # Replace the keys in kwargs with the substituted_non_default_args
            # so that the kwargs match the signature of the _forward_func
            kwargs = {k: v for k, v in zip(substituted_non_default_args, args)}

        # Create a partial domain geometry with the geometries corresponding
        # to the non-default arguments that are not in kwargs (remaining
        # unspecified inputs)
        partial_domain_geometry = cuqi.experimental.geometry._ProductGeometry(
            *[
                self.domain_geometry.geometries[i]
                for i in range(self.number_of_inputs)
                if original_non_default_args[i] not in kwargs.keys()
            ]
        )

        if len(partial_domain_geometry.geometries) == 1:
            partial_domain_geometry = partial_domain_geometry.geometries[0]

        # Create a domain geometry with the geometries corresponding to the
        # non-default arguments that are specified
        substituted_domain_geometry = cuqi.experimental.geometry._ProductGeometry(
            *[
                self.domain_geometry.geometries[i]
                for i in range(self.number_of_inputs)
                if original_non_default_args[i] in kwargs.keys()
            ]
        )

        if len(substituted_domain_geometry.geometries) == 1:
            substituted_domain_geometry = substituted_domain_geometry.geometries[0]

        # Create new model with partial input
        # First, we convert the input to function values
        kwargs = self._2fun(geometry=substituted_domain_geometry, is_par=is_par, **kwargs)

        # Second, we create a partial function for the forward operator
        partial_forward = partial(self._forward_func, **kwargs)

        # Third, if applicable, we create a partial function for the gradient
        if isinstance(self._gradient_func, tuple):
            # If gradient is a tuple, we create a partial function for each
            # gradient function in the tuple
            partial_gradient = tuple(
                (
                    partial(self._gradient_func[i], **kwargs)
                    if self._gradient_func[i] is not None
                    else None
                )
                for i in range(self.number_of_inputs)
                if original_non_default_args[i] not in kwargs.keys()
            )
            if len(partial_gradient) == 1:
                partial_gradient = partial_gradient[0]

        elif callable(self._gradient_func):
            raise NotImplementedError(
                "Partial forward model is only supported for gradient/jacobian functions that are tuples of callable functions."
            )

        else:
            partial_gradient = None

        # Lastly, we create the partial model with the partial forward
        # operator (we set the gradient function later)
        partial_model = Model(
            forward=partial_forward,
            range_geometry=self.range_geometry,
            domain_geometry=partial_domain_geometry,
        )

        # Set the _original_non_default_args (if applicable) and
        # _stored_non_default_args of the partial model
        if hasattr(self, "_original_non_default_args"):
            partial_model._original_non_default_args = reduced_original_non_default_args
        partial_model._stored_non_default_args = [
            self._non_default_args[i]
            for i in range(self.number_of_inputs)
            if original_non_default_args[i] not in kwargs.keys()
        ]

        # Set the gradient function of the partial model
        partial_model._check_correct_gradient_jacobian_form(
            partial_gradient, "gradient"
        )
        partial_model._gradient_func = partial_gradient

        return partial_model

    def _handle_case_when_model_input_is_distributions(self, kwargs):
        """Private function that handles the case of the input being a
        distribution or multiple distributions."""

        if not self._correct_distribution_dimension(kwargs.values()):
            raise ValueError(
                "Attempting to match parameter name of Model with given distribution(s), but distribution(s) dimension(s) does not match model input dimension(s)."
            )
        new_model = copy(self)

        # Store the original non_default_args of the model
        new_model._original_non_default_args = (
            self._original_non_default_args
            if hasattr(self, "_original_non_default_args")
            else self._non_default_args
        )

        # Update the non_default_args of the model to match the distribution
        # names. Defaults to x in the case of only one distribution that has no
        # name
        new_model._stored_non_default_args = [x.name for x in kwargs.values()]

        # If there is a repeated name, raise an error
        if len(set(new_model._stored_non_default_args)) != len(
            new_model._stored_non_default_args
        ):
            raise ValueError(
                "Attempting to match parameter name of Model with given distributions, but distribution names are not unique. Please provide unique names for the distributions."
            )

        return new_model

    def _handle_case_when_model_input_is_random_variables(self, kwargs):
        """ Private function that handles the case of the input being a random variable. """
        # If random variable is not a leaf-type node (e.g. internal node) we return NotImplemented
        if any(not isinstance(x.tree, cuqi.experimental.algebra.VariableNode) for x in kwargs.values()):
            return NotImplemented        

        # Extract the random variable distributions and check dimensions consistency with domain geometry
        distributions = [value.distribution for value in kwargs.values()]
        if not self._correct_distribution_dimension(distributions):
            raise ValueError("Attempting to match parameter name of Model with given random variable(s), but random variable dimension(s) does not match model input dimension(s).")

        new_model = copy(self)

        # Store the original non_default_args of the model
        new_model._original_non_default_args = self._non_default_args

        # Update the non_default_args of the model to match the random variable
        # names. Defaults to x in the case of only one random variable that has
        # no name
        new_model._stored_non_default_args = [x.name for x in distributions]

        # If there is a repeated name, raise an error
        if len(set(new_model._stored_non_default_args)) != len(
            new_model._stored_non_default_args
        ):
            raise ValueError(
                "Attempting to match parameter name of Model with given random variables, but random variables names are not unique. Please provide unique names for the random variables."
            )

        return new_model

    def gradient(
        self, direction, *args, is_direction_par=True, is_var_par=True, **kwargs
    ):
        """Gradient of the forward operator (Direction-Jacobian product)

        The gradient computes the Vector-Jacobian product (VJP) of the forward operator evaluated at the given model input and the given vector (direction).

        Parameters
        ----------
        direction : ndarray or cuqi.array.CUQIarray
            The direction at which to compute the gradient.

        *args : ndarrays or cuqi.array.CUQIarray objects
            Positional arguments for the values at which to compute the gradient. The gradient operator input can be specified as either positional arguments or keyword arguments but not both.

            If the input is specified as positional arguments, the order of the arguments should match the non_default_args of the model.

        is_direction_par : bool
            If True, `direction` is assumed to be parameters.
            If False, `direction` is assumed to be function values.

        is_var_par : bool or a tuple of bools
            If True, the inputs in `args` or `kwargs` are assumed to be parameters.
            If False, the inputs in `args` or `kwargs` are assumed to be function values.
            If `is_var_par` is a tuple of bools, the inputs in `args` or `kwargs` are assumed to be parameters or function values based on the corresponding boolean value in the tuple.
        """
        # Add args to kwargs and ensure the order of the arguments matches the
        # non_default_args of the forward function
        kwargs = self._parse_args_add_to_kwargs(
            *args, **kwargs, is_par=is_var_par, map_name="gradient"
        )

        # Obtain the parameters representation of the variables and raise an
        # error if it cannot be obtained
        error_message = (
            "For the gradient to be computed, is_var_par needs to be True and the variables in kwargs needs to be parameter value, not function value. Alternatively, the model domain_geometry:"
            + f" {self.domain_geometry} "
            + "should have an implementation of the method fun2par"
        )
        try:
            kwargs_par = self._2par(
                geometry=self.domain_geometry,
                is_par=is_var_par,
                to_CUQIarray=False,
                **kwargs,
            )
        # NotImplementedError will be raised if fun2par of the geometry is not
        # implemented and ValueError will be raised when imap is not set in
        # MappedGeometry
        except ValueError as e:
            raise ValueError(
                error_message
                + " ,including an implementation of imap for MappedGeometry"
            )
        except NotImplementedError as e:
            raise NotImplementedError(error_message)

        # Check for other errors that may prevent computing the gradient
        self._check_gradient_can_be_computed(direction, kwargs)

        # Also obtain the function values representation of the variables
        kwargs_fun = self._2fun(
            geometry=self.domain_geometry, is_par=is_var_par, **kwargs
        )

        # Store if any of the inputs is a CUQIarray
        to_CUQIarray = isinstance(direction, CUQIarray) or any(
            isinstance(x, CUQIarray) for x in kwargs_fun.values()
        )

        # Turn to_CUQIarray to a tuple of bools of the same length as kwargs_fun
        to_CUQIarray = tuple([to_CUQIarray] * len(kwargs_fun))

        # Convert direction to function value
        direction_fun = self._2fun(
            direction=direction, geometry=self.range_geometry, is_par=is_direction_par
        )

        # If model has _original_non_default_args, we use it to replace the
        # kwargs keys so that it matches self._gradient_func signature
        if hasattr(self, '_original_non_default_args'):
            args_fun = list(kwargs_fun.values())
            kwargs_fun = {
                k: v for k, v in zip(self._original_non_default_args, args_fun)
            }
        # Append the direction to the kwargs_fun as first input
        kwargs_fun_grad_input = {**direction_fun, **kwargs_fun}

        # Form 1 of gradient (callable)
        if callable(self._gradient_func):
            grad = self._gradient_func(**kwargs_fun_grad_input)
            grad_is_par = False  # Assume gradient is function value

        # Form 2 of gradient (tuple of callables)
        elif isinstance(self._gradient_func, tuple):
            grad = []
            for i, grad_func in enumerate(self._gradient_func):
                if grad_func is not None:
                    grad.append(grad_func(**kwargs_fun_grad_input))
                else:
                    grad.append(None)
                    # set the ith item of to_CUQIarray tuple to False
                    # because the ith gradient is None
                    to_CUQIarray = to_CUQIarray[:i] + (False,) + to_CUQIarray[i + 1 :]
            grad_is_par = False  # Assume gradient is function value

        grad = self._apply_chain_rule_to_account_for_domain_geometry_gradient(
            kwargs_par, grad, grad_is_par, to_CUQIarray
        )

        if len(grad) == 1:
            return list(grad.values())[0]
        elif self._gradient_output_stacked:
            return np.hstack(
                [
                    (
                        v.to_numpy()
                        if isinstance(v, CUQIarray)
                        else force_ndarray(v, flatten=True)
                    )
                    for v in list(grad.values())
                ]
            )

        return grad

    def _check_gradient_can_be_computed(self, direction, kwargs_dict):
        """Private function that checks if the gradient can be computed. By
        raising an error for the cases where the gradient cannot be computed."""

        # Raise an error if _gradient_func function is not set
        if self._gradient_func is None:
            raise NotImplementedError("Gradient is not implemented for this model.")

        # Raise an error if either the direction or kwargs are Samples objects
        if isinstance(direction, Samples) or any(
            isinstance(x, Samples) for x in kwargs_dict.values()
        ):
            raise NotImplementedError(
                "Gradient is not implemented for input of type Samples."
            )

        # Raise an error if range_geometry is not in the list returned by
        # `_get_identity_geometries()`. i.e. The Jacobian of its
        # par2fun map is not identity.
        # TODO: Add range geometry gradient to the chain rule
        if not type(self.range_geometry) in _get_identity_geometries():
            raise NotImplementedError(
                "Gradient is not implemented for model {} with range geometry {}. You can use one of the geometries in the list {}.".format(
                    self,
                    self.range_geometry,
                    [i_g.__name__ for i_g in _get_identity_geometries()],
                )
            )

        # Raise an error if domain_geometry (or its components in case of
        # _ProductGeometry) does not have gradient attribute and is not in the
        # list returned by `_get_identity_geometries()`. i.e. The Jacobian of its
        # par2fun map is not identity.
        domain_geometries = (
            self.domain_geometry.geometries
            if isinstance(
                self.domain_geometry, cuqi.experimental.geometry._ProductGeometry
            )
            else [self.domain_geometry]
        )
        for domain_geometry in domain_geometries:
            if (
                not hasattr(domain_geometry, "gradient")
                and not type(domain_geometry) in _get_identity_geometries()
            ):
                raise NotImplementedError(
                    "Gradient is not implemented for model \n{}\nwith domain geometry (or domain geometry component) {}. The domain geometries should have gradient method or be from the geometries in the list {}.".format(
                        self,
                        domain_geometry,
                        [i_g.__name__ for i_g in _get_identity_geometries()],
                    )
                )

    def _apply_chain_rule_to_account_for_domain_geometry_gradient(self,
                                                                  kwargs_par,
                                                                  grad,
                                                                  grad_is_par,
                                                                  to_CUQIarray):
        """ Private function that applies the chain rule to account for the
        gradient of the domain geometry. That is, it computes the gradient of
        the function values with respect to the parameters values."""
        # Create list of domain geometries
        geometries = (
            self.domain_geometry.geometries
            if isinstance(self.domain_geometry, cuqi.experimental.geometry._ProductGeometry)
            else [self.domain_geometry]
        )

        # turn grad_is_par to a tuple of bools if it is not already
        if isinstance(grad_is_par, bool):
            grad_is_par = tuple([grad_is_par]*self.number_of_inputs)

        # If the domain geometry is a _ProductGeometry and the gradient is
        # stacked, split it
        if (
            isinstance(
                self.domain_geometry, cuqi.experimental.geometry._ProductGeometry
            )
            and not isinstance(grad, (list, tuple))
            and isinstance(grad, np.ndarray)
        ):
            grad = np.split(grad, self.domain_geometry.stacked_par_split_indices)

        # If the domain geometry is not a _ProductGeometry, turn grad into a
        # list of length 1, so that we can iterate over it
        if not isinstance(self.domain_geometry, cuqi.experimental.geometry._ProductGeometry):
            grad = [grad]

        # apply the gradient of each geometry component
        grad_kwargs = {}
        for i, (k, v_par) in enumerate(kwargs_par.items()):
            if hasattr(geometries[i], 'gradient') and grad[i] is not None:
                grad_kwargs[k] = geometries[i].gradient(grad[i], v_par)
                # update the ith component of grad_is_par to True
                grad_is_par = grad_is_par[:i] + (True,) + grad_is_par[i+1:]
            else:
                grad_kwargs[k] = grad[i]

        # convert the computed gradient to parameters
        grad = self._2par(geometry=self.domain_geometry,
                          to_CUQIarray=to_CUQIarray,
                          is_par=grad_is_par,
                          **grad_kwargs)

        return grad

    def __call__(self, *args, **kwargs):
        return self.forward(*args, **kwargs)

    def __len__(self):
        return self.range_dim

    def __repr__(self) -> str:
        kwargs = {}
        if self.number_of_inputs > 1:
            pad = " " * len("CUQI {}: ".format(self.__class__.__name__))
            kwargs["pad"]=pad  
        return "CUQI {}: {} -> {}.\n    Forward parameters: {}.".format(self.__class__.__name__,self.domain_geometry.__repr__(**kwargs),self.range_geometry,self._non_default_args)

class AffineModel(Model):
    """ Model class representing an affine model, i.e. a linear operator with a fixed shift. For linear models, represented by a linear operator only, see :class:`~cuqi.model.LinearModel`.

    The affine model is defined as:

    .. math::

        x \\mapsto Ax + shift

    where :math:`A` is the linear operator and :math:`shift` is the shift.

    Parameters
    ----------

    linear_operator : 2d ndarray, callable function or cuqi.model.LinearModel
        The linear operator. If ndarray is given, the operator is assumed to be a matrix.

    shift : scalar or array_like
        The shift to be added to the forward operator.

    linear_operator_adjoint : callable function, optional
        The adjoint of the linear operator. Also used for computing gradients.

    range_geometry : cuqi.geometry.Geometry
        The geometry representing the range.

    domain_geometry : cuqi.geometry.Geometry
        The geometry representing the domain.

    """

    def __init__(self, linear_operator, shift, linear_operator_adjoint=None, range_geometry=None, domain_geometry=None):

        # If input represents a matrix, extract needed properties from it
        if hasattr(linear_operator, '__matmul__') and hasattr(linear_operator, 'T'):
            if linear_operator_adjoint is not None:
                raise ValueError("Adjoint of linear operator should not be provided when linear operator is a matrix. If you want to provide an adjoint, use a callable function for the linear operator.")

            matrix = linear_operator

            linear_operator = lambda x: matrix@x
            linear_operator_adjoint = lambda y: matrix.T@y

            if range_geometry is None:
                if hasattr(matrix, 'shape'):
                    range_geometry = _DefaultGeometry1D(grid=matrix.shape[0])
                elif isinstance(matrix, LinearModel):
                    range_geometry = matrix.range_geometry

            if domain_geometry is None:
                if hasattr(matrix, 'shape'):
                    domain_geometry = _DefaultGeometry1D(grid=matrix.shape[1])
                elif isinstance(matrix, LinearModel):
                    domain_geometry = matrix.domain_geometry
        else:
            matrix = None

        # Ensure that the operators are a callable functions (either provided or created from matrix)
        if not callable(linear_operator):
            raise TypeError("Linear operator must be defined as a matrix or a callable function of some kind")
        if linear_operator_adjoint is not None and not callable(linear_operator_adjoint):
            raise TypeError("Linear operator adjoint must be defined as a callable function of some kind")

        # If linear operator is of type Model, it needs to be a LinearModel
        if isinstance(linear_operator, Model) and not isinstance(
            linear_operator, LinearModel
        ):
            raise TypeError(
                "The linear operator should be a LinearModel object, a callable function or a matrix."
            )

        # If the adjoint operator is of type Model, it needs to be a LinearModel
        if isinstance(linear_operator_adjoint, Model) and not isinstance(
            linear_operator_adjoint, LinearModel
        ):
            raise TypeError(
                "The adjoint linear operator should be a LinearModel object, a callable function or a matrix."
            )

        # Additional checks if the linear_operator is not a LinearModel:
        if not isinstance(linear_operator, LinearModel):
            # Ensure the linear operator has exactly one input argument
            if len(cuqi.utilities.get_non_default_args(linear_operator)) != 1:
                raise ValueError(
                    "The linear operator should have exactly one input argument."
                )
            # Ensure the adjoint linear operator has exactly one input argument
            if (
                linear_operator_adjoint is not None
                and len(cuqi.utilities.get_non_default_args(linear_operator_adjoint))
                != 1
            ):
                raise ValueError(
                    "The adjoint linear operator should have exactly one input argument."
                )

        # Check size of shift and match against range_geometry
        if not np.isscalar(shift):
            if len(shift) != range_geometry.par_dim:
                raise ValueError("The shift should have the same dimension as the range geometry.")

        # Store linear operator privately
        # Note: we need to set the _linear_operator before calling the
        # super().__init__() because it is needed when calling the property
        # _non_default_args within the super().__init__()
        self._linear_operator = linear_operator

        # Initialize Model class
        super().__init__(linear_operator, range_geometry, domain_geometry)

        # Store matrix privately
        self._matrix = matrix

        # Store shift as private attribute
        self._shift = shift


        # Store adjoint function
        self._linear_operator_adjoint = linear_operator_adjoint

        # Define gradient
        self._gradient_func = lambda direction, *args, **kwargs: linear_operator_adjoint(direction)

        # Update forward function to include shift (overwriting the one from Model class)
        self._forward_func = lambda *args, **kwargs: linear_operator(*args, **kwargs) + shift

        # Set stored_non_default_args to None
        self._stored_non_default_args = None

    @property
    def _non_default_args(self):
        if self._stored_non_default_args is None:
            # Use arguments from user's callable linear operator
            self._stored_non_default_args = cuqi.utilities.get_non_default_args(
                self._linear_operator
            )
        return self._stored_non_default_args

    @property
    def shift(self):
        """ The shift of the affine model. """
        return self._shift

    @shift.setter
    def shift(self, value):
        """ Update the shift of the affine model. Updates both the shift value and the underlying forward function. """
        self._shift = value
        self._forward_func = lambda *args, **kwargs: self._linear_operator(*args, **kwargs) + value

    def _forward_func_no_shift(self, *args, is_par=True, **kwargs):
        """Helper function for computing the forward operator without the shift."""
        # convert args to kwargs
        kwargs = self._parse_args_add_to_kwargs(
            *args, **kwargs, map_name="model", is_par=is_par
        )
        args = list(kwargs.values())
        # if model has _original_non_default_args, we use it to replace the
        # kwargs keys so that it matches self._linear_operator signature
        if hasattr(self, '_original_non_default_args'):
            kwargs = {k:v for k,v in zip(self._original_non_default_args, args)}
        return self._apply_func(self._linear_operator, **kwargs, is_par=is_par)

    def _adjoint_func_no_shift(self, *args, is_par=True, **kwargs):
        """Helper function for computing the adjoint operator without the shift."""
        # convert args to kwargs
        kwargs = self._parse_args_add_to_kwargs(
            *args,
            **kwargs,
            map_name='adjoint',
            is_par=is_par,
            non_default_args=cuqi.utilities.get_non_default_args(
                self._linear_operator_adjoint
            ),
        )
        return self._apply_func(
            self._linear_operator_adjoint, **kwargs, is_par=is_par, fwd=False
        )


class LinearModel(AffineModel):
    """Model based on a Linear forward operator.

    Parameters
    -----------
    forward : 2D ndarray or callable function.
        Forward operator.

    adjoint : 2D ndarray or callable function. (optional if matrix is passed as forward)

    range_geometry : integer or cuqi.geometry.Geometry (optional)
        If integer is given, a cuqi.geometry._DefaultGeometry is created with dimension of the integer.

    domain_geometry : integer or cuqi.geometry.Geometry (optional)
        If integer is given, a cuqi.geometry._DefaultGeometry is created with dimension of the integer.


    :ivar range_geometry: The geometry representing the range.
    :ivar domain_geometry: The geometry representing the domain.

    Example
    -------

    Consider a linear model represented by a matrix, i.e., :math:`y=Ax` where 
    :math:`A` is a matrix.
    
    We can define such a linear model by passing the matrix :math:`A`:

    .. code-block:: python

        import numpy as np
        from cuqi.model import LinearModel

        A = np.random.randn(2,3)

        model = LinearModel(A)

    The dimension of the range and domain geometries will be automatically 
    inferred from the matrix :math:`A`.
        
    Meanwhile, such a linear model can also be defined by a forward function 
    and an adjoint function:

    .. code-block:: python

        import numpy as np
        from cuqi.model import LinearModel

        A = np.random.randn(2,3)

        def forward(x):
            return A@x

        def adjoint(y):
            return A.T@y

        model = LinearModel(forward,
                            adjoint=adjoint,
                            range_geometry=2,
                            domain_geometry=3)

    Note that you would need to specify the range and domain geometries in this
    case as they cannot be inferred from the forward and adjoint functions.
    """

    def __init__(self, forward, adjoint=None, range_geometry=None, domain_geometry=None):

        # Initialize as AffineModel with shift=0
        super().__init__(forward, 0, adjoint, range_geometry, domain_geometry)

    def adjoint(self, *args, is_par=True, **kwargs):
        """ Adjoint of the model.
        
        Adjoint converts the input to function values (if needed) using the range geometry of the model then applies the adjoint operator to the function values and converts the output function values to parameters using the domain geometry of the model.

        Parameters
        ----------
        *args : ndarrays or cuqi.array.CUQIarray object
            Positional arguments for the adjoint operator ( maximum one argument). The adjoint operator input can be specified as either positional arguments or keyword arguments but not both.

        **kwargs : keyword arguments
            keyword arguments for the adjoint operator (maximum one argument). The adjoint operator input can be specified as either positional arguments or keyword arguments but not both.

            If the input is specified as keyword arguments, the keys should match the non_default_args of the model.

        Returns
        -------
        ndarray or cuqi.array.CUQIarray
            The adjoint model output. Always returned as parameters.
        """
        kwargs = self._parse_args_add_to_kwargs(
            *args,
            **kwargs,
            map_name='adjoint',
            is_par=is_par,
            non_default_args=cuqi.utilities.get_non_default_args(
                self._linear_operator_adjoint
            ),
        )

        # length of kwargs should be 1
        if len(kwargs) > 1:
            raise ValueError(
                "The adjoint operator input is specified by more than one argument. This is not supported."
            )
        if self._linear_operator_adjoint is None:
            raise ValueError("No adjoint operator was provided for this model.")
        return self._apply_func(
            self._linear_operator_adjoint, **kwargs, is_par=is_par, fwd=False
        )

    def __matmul__(self, *args, **kwargs):
        return self.forward(*args, **kwargs)

    def get_matrix(self):
        """
        Returns an ndarray with the matrix representing the forward operator.
        """
        if self._matrix is not None: #Matrix exists so return it
            return self._matrix
        else:
            # TODO: Can we compute this faster while still in sparse format?
            mat = csc_matrix((self.range_dim,0)) #Sparse (m x 1 matrix)
            e = np.zeros(self.domain_dim)

            # Stacks sparse matrices on csc matrix
            for i in range(self.domain_dim):
                e[i] = 1
                col_vec = self.forward(e)
                mat = hstack((mat,col_vec[:,None])) #mat[:,i] = self.forward(e)
                e[i] = 0

            # Store matrix for future use
            self._matrix = mat

            return self._matrix   

    @property
    def T(self):
        """Transpose of linear model. Returns a new linear model acting as the transpose."""
        transpose = LinearModel(
            self._linear_operator_adjoint,
            self._linear_operator,
            self.domain_geometry,
            self.range_geometry,
        )
        if self._matrix is not None:
            transpose._matrix = self._matrix.T
        return transpose


class PDEModel(Model):
    """
    Model based on an underlying cuqi.pde.PDE.
    In the forward method the PDE is assembled, solved and observed.
    
    Parameters
    -----------
    PDE : cuqi.pde.PDE
        The PDE that specifies the forward operator.

    range_geometry : integer or cuqi.geometry.Geometry, optional
        If integer is given, a cuqi.geometry._DefaultGeometry is created with dimension of the integer.

    domain_geometry : integer or cuqi.geometry.Geometry, optional
        If integer is given, a cuqi.geometry._DefaultGeometry is created with dimension of the integer.


    :ivar range_geometry: The geometry representing the range.
    :ivar domain_geometry: The geometry representing the domain.
    """
    def __init__(self, PDE: cuqi.pde.PDE, range_geometry, domain_geometry, **kwargs):

        if not isinstance(PDE, cuqi.pde.PDE):
            raise ValueError("PDE needs to be a cuqi PDE.")
        # PDE needs to be set before calling super().__init__
        # for the property _non_default_args to work
        self.pde = PDE
        self._stored_non_default_args = None

        # If gradient or jacobian is not provided, we create it from the PDE
        if not np.any([k in kwargs.keys() for k in ["gradient", "jacobian"]]):
            # Create gradient or jacobian function to pass to the Model based on
            # the PDE object. The dictionary derivative_kwarg contains the
            # created function along with the function type (either "gradient"
            # or "jacobian")
            derivative_kwarg = self._create_derivative_function()
            # append derivative_kwarg to kwargs
            kwargs.update(derivative_kwarg)

        super().__init__(forward=self._forward_func_pde,
                         range_geometry=range_geometry,
                         domain_geometry=domain_geometry,
                         **kwargs)

    @property
    def _non_default_args(self):
        if self._stored_non_default_args is None:
            # extract the non-default arguments of the PDE
            self._stored_non_default_args = self.pde._non_default_args

        return self._stored_non_default_args

    def _forward_func_pde(self, **kwargs):

        self.pde.assemble(**kwargs)

        sol, info = self.pde.solve()

        obs = self.pde.observe(sol)

        return obs

    def _create_derivative_function(self):
        """Private function that creates the derivative function (gradient or
        jacobian) based on the PDE object. The derivative function is created as
        a lambda function that takes the direction and the parameters as input 
        and returns the gradient or jacobian of the PDE. This private function
        returns a dictionary with the created function and the function type
        (either "gradient" or "jacobian")."""

        if hasattr(self.pde, "gradient_wrt_parameter"):
            # Build the string that will be used to create the lambda function
            function_str = (
                "lambda direction, "
                + ", ".join(self._non_default_args)
                + ", pde_func: pde_func(direction, "
                + ", ".join(self._non_default_args)
                + ")"
            )

            # create the lambda function from the string
            function = eval(function_str)

            # create partial function from the lambda function with gradient_wrt_parameter
            # as the first argument
            grad_func = partial(function, pde_func=self.pde.gradient_wrt_parameter)

            # Return the gradient function
            return {"gradient": grad_func}

        elif hasattr(self.pde, "jacobian_wrt_parameter"):
            # Build the string that will be used to create the lambda function
            function_str = (
                "lambda "
                + ", ".join(self._non_default_args)
                + ", pde_func: pde_func( "
                + ", ".join(self._non_default_args)
                + ")"
            )

            # create the lambda function from the string
            function = eval(function_str)

            # create partial function from the lambda function with jacobian_wrt_parameter
            # as the first argument
            jacobian_func = partial(function, pde_func=self.pde.jacobian_wrt_parameter)

            # Return the jacobian function
            return {"jacobian": jacobian_func}
        else:
            return {} # empty dictionary if no gradient or jacobian is found

    # Add the underlying PDE class name to the repr.
    def __repr__(self) -> str:
        return super().__repr__()+"\n    PDE: {}.".format(self.pde.__class__.__name__)



import numpy as np
from numpy import linalg as LA
from scipy.optimize import fmin_l_bfgs_b, least_squares
import scipy.optimize as opt
import scipy.sparse as spa

from cuqi.array import CUQIarray
from cuqi import config
eps = np.finfo(float).eps

try:
    from sksparse.cholmod import cholesky
    has_cholmod = True
except ImportError:
    has_cholmod = False


class ScipyLBFGSB(object):
    """Wrapper for :meth:`scipy.optimize.fmin_l_bfgs_b`.

    Minimize a function func using the L-BFGS-B algorithm.
    
    Note, Scipy does not recommend using this method.

    Parameters
    ----------
    func : callable f(x,*args)
        Function to minimize.
    x0 : ndarray
        Initial guess.
    gradfunc : callable f(x,*args), optional
        The gradient of func. 
        If None, the solver approximates the gradient with a finite difference scheme.
    kwargs : keyword arguments passed to scipy's L-BFGS-B algorithm. See documentation for scipy.optimize.minimize
    """
    def __init__(self, func, x0, gradfunc = None, **kwargs):
        self.func= func
        self.x0 = x0
        self.gradfunc = gradfunc
        self.kwargs = kwargs
    
    def solve(self):
        """Runs optimization algorithm and returns solution and info.

        Returns
        ----------
        solution : array_like
            Estimated position of the minimum.
        info : dict
            Information dictionary.
            success: 1 if minimization has converged, 0 if not.
            message: Description of the cause of the termination.
            func: Function value at the estimated minimum.
            grad: Gradient at the estimated minimum.
            nit: Number of iterations.
            nfev: Number of func evaluations.
        """
        # Check if there is a gradient. If not, let the solver use an approximate gradient
        if self.gradfunc is None:
            approx_grad = 1
        else:
            approx_grad = 0
        # run solver
        solution = fmin_l_bfgs_b(self.func,self.x0, fprime = self.gradfunc, approx_grad = approx_grad, **self.kwargs)
        if solution[2]['warnflag'] == 0:
            success = 1
            message = 'Optimization terminated successfully.'
        elif solution[2]['warnflag'] == 1:
            success = 0
            message = 'Terminated due to too many function evaluations or too many iterations.'
        else:
            success = 0
            message = solution[2]['task']
        info = {"success": success,
                "message": message,
                "func": solution[1],
                "grad": solution[2]['grad'],
                "nit": solution[2]['nit'], 
                "nfev": solution[2]['funcalls']}
        return solution[0], info

class ScipyMinimizer(object):
    """Wrapper for :meth:`scipy.optimize.minimize`.

    Minimize a function func using scipy's optimize.minimize module.
    
    Parameters
    ----------
    func : callable f(x,*args)
        Function to minimize.
    x0 : ndarray
        Initial guess.
    gradfunc : callable f(x,*args), optional
        The gradient of func. 
        If None, then the solver approximates the gradient.
    method : str or callable, optional
        Type of solver. Should be one of
        ‘Nelder-Mead’
        ‘Powell’
        ‘CG’
        ‘BFGS’
        ‘Newton-CG’ 
        ‘L-BFGS-B’
        ‘TNC’ 
        ‘COBYLA’ 
        ‘SLSQP’
        ‘trust-constr’
        ‘dogleg’ 
        ‘trust-ncg’ 
        ‘trust-exact’ 
        ‘trust-krylov’ 
        If not given, chosen to be one of BFGS, L-BFGS-B, SLSQP, depending if the problem has constraints or bounds.
    kwargs : keyword arguments passed to scipy's minimizer. See documentation for scipy.optimize.minimize
    """
    def __init__(self, func, x0, gradfunc = '2-point', method = None, **kwargs):
        self.func= func
        self.x0 = x0
        self.method = method
        self.gradfunc = gradfunc
        self.kwargs = kwargs
    
    def solve(self):
        """Runs optimization algorithm and returns solution and info.

        Returns
        ----------
        solution : array_like
            Estimated position of the minimum.
        info : dict
            Information dictionary.
            success: 1 if minimization has converged, 0 if not.
            message: Description of the cause of the termination.
            func: Function value at the estimated minimum.
            grad: Gradient at the estimated minimum.
            nit: Number of iterations.
            nfev: Number of func evaluations.
        """
        solution = opt.minimize(self.func, self.x0, jac = self.gradfunc, method = self.method, **self.kwargs)
        info = {"success": solution['success'],
                "message": solution['message'],
                "func": solution['fun'],
                "nit": solution['nit'], 
                "nfev": solution['nfev']}
        # if gradfunc is callable, record the gradient in the info dict
        if 'jac' in solution.keys():
            info['grad'] = solution['jac']
        if isinstance(self.x0,CUQIarray):
            sol = CUQIarray(solution['x'],geometry=self.x0.geometry)
        else:
            sol = solution['x']
        return sol, info

class ScipyMaximizer(ScipyMinimizer):
    """Simply calls ::class:: cuqi.solver.ScipyMinimizer with -func."""
    def __init__(self, func, x0, gradfunc = None, method = None, **kwargs):
        def nfunc(*args,**kwargs):
            return -func(*args,**kwargs)
        if gradfunc is not None:
            def ngradfunc(*args,**kwargs):
                return -gradfunc(*args,**kwargs)
        else:
            ngradfunc = gradfunc
        super().__init__(nfunc,x0,ngradfunc,method,**kwargs)



class ScipyLSQ(object):
    """Wrapper for :meth:`scipy.optimize.least_squares`.

    Solve nonlinear least-squares problems with bounds:

    .. math::
    
        \min F(x) = 0.5 * \sum(\\rho(f_i(x)^2), i = 0, ..., m-1)

    subject to :math:`lb <= x <= ub`.
    
    Parameters
    ----------
    func : callable f(x,*args)
        Function to minimize.
    x0 : ndarray
        Initial guess.
    Jac : callable f(x,*args), optional
        The Jacobian of func. 
        If not specified, the solver approximates the Jacobian with a finite difference scheme.
    loss: callable rho(x,*args)
        Determines the loss function
        'linear' : rho(z) = z. Gives a standard least-squares problem.
        'soft_l1': rho(z) = 2*((1 + z)**0.5 - 1). The smooth approximation of l1 (absolute value) loss.
        'huber'  : rho(z) = z if z <= 1 else 2*z**0.5 - 1. Works similar to 'soft_l1'.
        'cauchy' : rho(z) = ln(1 + z). Severely weakens outliers influence, but may cause difficulties in optimization process.
        'arctan' : rho(z) = arctan(z). Limits a maximum loss on a single residual, has properties similar to 'cauchy'.        
    method : str or callable, optional
        Type of solver. Should be one of
        'trf', Trust Region Reflective algorithm: for large sparse problems with bounds.
        'dogbox', dogleg algorithm with rectangular trust regions, for small problems with bounds.
        'lm', Levenberg-Marquardt algorithm as implemented in MINPACK. Doesn't handle bounds and sparse Jacobians.
    tol : The numerical tolerance for convergence checks.
    maxit : The maximum number of iterations.
    kwargs : Additional keyword arguments passed to scipy's least_squares. Empty by default. See documentation for scipy.optimize.least_squares
    """
    def __init__(self, func, x0, jacfun='2-point', method='trf', loss='linear', tol=1e-6, maxit=1e4, **kwargs):
        self.func = func
        self.x0 = x0
        self.jacfun = jacfun
        self.method = method
        self.loss = loss
        self.tol = tol
        self.maxit = int(maxit)
        self.kwargs = kwargs
    
    def solve(self):
        """Runs optimization algorithm and returns solution and info.

        Returns
        ----------
        solution : Tuple
            Solution found (array_like) and optimization information (dictionary).
        """
        solution = least_squares(self.func, self.x0, jac=self.jacfun, \
                                method=self.method, loss=self.loss, xtol=self.tol, max_nfev=self.maxit, **self.kwargs)
        info = {"success": solution['success'],
                "message": solution['message'],
                "func": solution['fun'],
                "jac": solution['jac'],
                "nfev": solution['nfev']}
        if isinstance(self.x0, CUQIarray):
            sol = CUQIarray(solution['x'], geometry=self.x0.geometry)
        else:
            sol = solution['x']
        return sol, info

class ScipyLinearLSQ(object):
    """Wrapper for :meth:`scipy.optimize.lsq_linear`.

    Solve linear least-squares problems with bounds:

    .. math::
    
        \min \|A x - b\|_2^2

    subject to :math:`lb <= x <= ub`.
    
    Parameters
    ----------
    A : ndarray, LinearOperator
        Design matrix (system matrix).
    b : ndarray
        The right-hand side of the linear system.
    bounds : 2-tuple of array_like or scipy.optimize Bounds
        Bounds for variables. 
    kwargs : Other keyword arguments passed to Scipy's `lsq_linear`. See documentation of `scipy.optimize.lsq_linear` for details.
    """
    def __init__(self, A, b, bounds=(-np.inf, np.inf), **kwargs):
        self.A = A
        self.b = b
        self.bounds = bounds
        self.kwargs = kwargs
    
    def solve(self):
        """Runs optimization algorithm and returns solution and optimization information.

        Returns
        ----------
        solution : Tuple
            Solution found (array_like) and optimization information (dictionary).
        """
        res = opt.lsq_linear(self.A, self.b, bounds=self.bounds, **self.kwargs)
        x = res.pop('x')
        return x, res


class CGLS(object):
    """Conjugate Gradient method for unsymmetric linear equations and least squares problems.

    See http://web.stanford.edu/group/SOL/software/cgls/ for the matlab version it is based on.
    
    If SHIFT is 0, then CGLS is Hestenes and Stiefel's conjugate-gradient method for least-squares problems. 
    If SHIFT is nonzero, the system (A'*A + SHIFT*I)*X = A'*b is solved.

    Solve Ax=b or minimize ||Ax-b||^2 or solve (A^TA+sI)x=A^Tb.
    
    Parameters
    ----------
    A : ndarray or callable f(x,*args).
    b : ndarray.
    x0 : ndarray. Initial guess.
    maxit : The maximum number of iterations.
    tol : The numerical tolerance for convergence checks.
    shift : The shift parameter (s) shown above.
    """    
    def __init__(self, A, b, x0, maxit, tol=1e-6, shift=0):
        self.A = A
        self.b = b
        self.x0 = x0
        self.maxit = int(maxit)
        self.tol = tol        
        self.shift = shift
        if not callable(A):
            self.explicitA = True
        else:
            self.explicitA = False
            
    def solve(self):
        # initial state
        x = self.x0.copy()
        if self.explicitA:
            r = self.b - (self.A @ x)
            s = (self.A.T @ r) - self.shift*x
        else:        
            r = self.b - self.A(x, 1)
            s = self.A(r, 2) - self.shift*x
    
        # initialization
        p = s.copy()
        norms0 = LA.norm(s)
        normx = LA.norm(x)
        gamma, xmax = norms0**2, normx
    
        # main loop
        k, flag, indefinite = 0, 0, 0
        while (k < self.maxit) and (flag == 0):
            k += 1
            if self.explicitA:
                q = self.A @ p
            else:
                q = self.A(p, 1)
            delta_cgls = LA.norm(q)**2 + self.shift*LA.norm(p)**2

            if (delta_cgls < 0):
                indefinite = True
            elif (delta_cgls == 0):
                delta_cgls = eps
            alpha_cgls = gamma / delta_cgls

            x += alpha_cgls*p
            r -= alpha_cgls*q
            if self.explicitA:
                s = self.A.T @ r - self.shift*x     
            else:
                s = self.A(r, 2) - self.shift*x

            gamma1 = gamma.copy()
            norms = LA.norm(s)
            gamma = norms**2
            p = s + (gamma/gamma1)*p
        
            # convergence
            normx = LA.norm(x)
            xmax = max(xmax, normx)
            flag = (norms <= norms0*self.tol) or (normx*self.tol >= 1)
            # resNE = norms / norms0

        shrink = normx/xmax
        # if k == self.maxit:          
        #     flag = 2   # CGLS iterated MAXIT times but did not converge
        #     Warning('\n maxit reached without convergence !')
        if indefinite:          
            flag = 3   # Matrix (A'*A + delta*L) seems to be singular or indefinite
            ValueError('\n Negative curvature detected !')  
        if shrink <= np.sqrt(self.tol):
            flag = 4   # Instability likely: (A'*A + delta*L) indefinite and NORM(X) decreased
            ValueError('\n Instability likely !') 
    
        return x, k


class PCGLS:
    """Conjugate Gradient method for least squares problems with preconditioning
    
    See Bjorck (1996) - Numerical Methods for Least Squares Problems. Pag 294.

    Parameters
    ----------
    A : ndarray or callable f(x,*args).
        Function or array representing the forward model.
    b : ndarray
        Data vector.
    x0 : ndarray
        Initial guess.    
    P : ndarray
        Preconditioner array in sparse format.
    maxit : int
        The maximum number of iterations.
    tol : float
        The numerical tolerance for convergence checks.
    """    
    def __init__(self, A, b, x0, P, maxit, tol=1e-6, shift=0):
        self._A = A
        self._b = b
        self._x0 = x0
        self._P = P
        self._maxit = int(maxit)
        self._tol = tol        
        self._shift = shift
        self._dim = len(x0)
        if not callable(A):
            self._explicitA = True
        else:
            self._explicitA = False
        #
        if self._dim < config.MAX_DIM_INV:
            self._explicitPinv = True
            Pinv = spa.linalg.inv(P)
        else:
            self._explicitPinv = False
            Pinv = None # we do cholesky.solve or sparse.solve with P
            if has_cholmod:
                # turn P into a cholesky object
                P = cholesky(P, ordering_method='natural')
        self._Pinv = Pinv # inverse of the preconditioner as a ndarray or function

    def solve(self):
        # initial state
        x = self._x0.copy()
        r = self._b - self._apply_A(x, 1)
        s = self._apply_Pinv(self._apply_A(r, 2), 2)
        p = s.copy()

        # initial computations        
        norms0 = LA.norm(s)
        normx = LA.norm(x)
        gamma, xmax = norms0**2, normx

        # main loop
        k, flag, indefinite = 0, 0, 0
        while (k < self._maxit) and (flag == 0):
            k += 1
            #
            t = self._apply_Pinv(p, 1)
            q = self._apply_A(t, 1)
            #
            delta_cgls = LA.norm(q)**2
            if (delta_cgls < 0):
                indefinite = True
            elif (delta_cgls == 0):
                delta_cgls = eps
            alpha_cgls = gamma / delta_cgls
            #
            x += alpha_cgls*t
            r -= alpha_cgls*q
            s = self._apply_Pinv(self._apply_A(r, 2), 2)
            #
            norms = LA.norm(s)
            gamma1 = gamma.copy()
            gamma = norms**2
            beta = gamma / gamma1
            p = s + beta*p

            # convergence
            normx = LA.norm(x)
            xmax = max(xmax, normx)
            flag = (norms <= norms0*self._tol) or (normx*self._tol >= 1)
            # resNE = norms / norms0

        shrink = normx/xmax
        if indefinite:          
            flag = 3   # Matrix (A'*A + delta*L) seems to be singular or indefinite
            ValueError('\n Negative curvature detected !')  
        if shrink <= np.sqrt(self._tol):
            flag = 4   # Instability likely: (A'*A + delta*L) indefinite and NORM(X) decreased
            ValueError('\n Instability likely !') 

        return x, k

    def _apply_A(self, x, flag):
        # applies system operator A: forward or adjoint
        if self._explicitA:
            if flag == 1:
                evalu = self._A @ x
            elif flag == 2:
                evalu = self._A.T @ x
        else:
            evalu = self._A(x, flag)
        return evalu

    def _apply_Pinv(self, x, flag):
        # applies the inverse of the preconditioner P: forward or adjoint (see Bjorck (1996) P. 294)
        if self._explicitPinv:
            if flag == 1:
                precond = self._Pinv @ x
            elif flag == 2:
                 precond = self._Pinv.T @ x
        else:
            if has_cholmod:
                if flag == 1:
                    precond = self._P.solve_A(x, use_LDLt_decomposition=False) 
                elif flag == 2:
                    precond = self._P.solve_At(x, use_LDLt_decomposition=False) 
            else:
                if flag == 1:
                    precond = spa.linalg.spsolve(self._P, x) 
                elif flag == 2:
                    precond = spa.linalg.spsolve(self._P.T, x)
        return precond



class LM(object):
    """Levenberg-Marquardt algorithm for nonlinear least-squares problems.
    This is a translation of LevMaq.m from
    https://github.com/bardsleyj/SIAMBookCodes/tree/master/Functions    
    Used in Bardsley (2019) - Computational UQ for inverse problems. SIAM.

    Based Algorithm 3.3.5 from:
    Kelley (1999) - Iterative Methods for Optimization. SIAM.

    Parameters
    ----------
    A : callable f(x,*args).
    x0 : ndarray. Initial guess.
    jacfun : callable Jac(x). Jacobian of func. 
    maxit : The maximum number of iterations.
    tol : The numerical tolerance for convergence checks.
    gradtol : The numerical tolerance for gradient.
    nu0 : default value for nu parameter in the algorithm.
    sparse : True: if 'jacfun' is defined as a sparse matrix, or as a function that 
            supports sparse operations. False: if 'jacfun' is dense.
    """
    def __init__(self, A, x0, jacfun, maxit=1e4, tol=1e-6, gradtol=1e-8, nu0=1e-3, sparse=True):
        self.A = A
        self.x0 = x0
        self.jacfun = jacfun
        self.maxit = int(maxit)
        self.tol = tol
        self.gradtol = gradtol
        self.nu0 = nu0
        self.n = len(x0)
        self.sparse = sparse
        if not callable(A): # also applies for jacfun
            self.explicitA = True
        else:
            self.explicitA = False

    def solve(self):
        x = self.x0
        if self.explicitA:
            r = self.A @ x
            J = self.jacfun @ x
        else:
            r = self.A(x)
            J = self.jacfun(x)
        g = J.T @ r
        ng = LA.norm(g)
        ng0, nu = np.copy(ng), np.copy(ng)
        f = 0.5*(r.T @ r)
        i = 0

        # solve function depending on sparsity
        if self.sparse:
            insolve = lambda A, b: spa.linalg.spsolve(A, b)
            I = spa.identity(self.n)
        else:
            insolve = lambda A, b: LA.solve(A, b)
            I = np.identity(self.n)

        # parameters required for the LM parameter update
        mu0, mulow, muhigh = 0, 0.25, 0.75
        omdown, omup = 0.5, 2
        while ((ng/ng0) > self.gradtol) and (i < self.maxit):
            i += 1
            s = insolve((J.T@J + nu*I), g)
            xtemp = x-s
            if self.explicitA:
                rtemp = self.A @ xtemp
                Jtemp = self.jacfun @ xtemp
            else:
                rtemp = self.A(xtemp)
                Jtemp = self.jacfun(xtemp)
            ftemp = 0.5*(rtemp.T @ rtemp)

            # LM parameter update
            num, den = f-ftemp, (xtemp-x).T @ g
            if (num != 0) and (den != 0): 
                ratio = -2*(num / den)
            else:
                ratio = 0
            if (ratio < mu0):
                nu = max(omup*nu, self.nu0)
            else:
                x, r, f = np.copy(xtemp), np.copy(rtemp), np.copy(ftemp)
                if self.sparse:
                    J = spa.csr_matrix.copy(Jtemp)
                else:
                    J = np.copy(Jtemp)
                if (ratio < mulow):
                    nu = max(omup*nu, self.nu0)
                elif (ratio > muhigh):
                    nu = omdown*nu
                    if (nu < self.nu0):
                        nu = 0
            g = J.T @ r
            ng = LA.norm(g)

        info = {"func": r,
                "Jac": J,
                "nfev": i}
        return x, info



class PDHG(object):
    """Primal-Dual Hybrid Gradient algorithm."""
    def __init__(self):
        raise NotImplementedError



class FISTA(object):
    """Fast Iterative Shrinkage-Thresholding Algorithm for regularized least squares problems.
    
    Reference:
    Beck, Amir, and Marc Teboulle. "A fast iterative shrinkage-thresholding algorithm for linear inverse problems." SIAM journal on imaging sciences 2.1 (2009): 183-202.

    Minimize ||Ax-b||^2 + f(x).
    
    
    Parameters
    ----------
    A : ndarray or callable f(x,*args).
    b : ndarray.
    proximal : callable f(x, gamma) for proximal mapping.
    x0 : ndarray. Initial guess.
    maxit : The maximum number of iterations.
    stepsize : The stepsize of the gradient step.
    abstol : The numerical tolerance for convergence checks.
    adapative : Whether to use FISTA or ISTA.

    Example
    -----------
    .. code-block:: python
    
        from cuqi.solver import FISTA,  ProximalL1
        import scipy as sp
        import numpy as np

        rng = np.random.default_rng()

        m, n = 10, 5
        A = rng.standard_normal((m, n))
        b = rng.standard_normal(m)
        stepsize = 0.99/(sp.linalg.interpolative.estimate_spectral_norm(A)**2)
        x0 = np.zeros(n)
        fista = FISTA(A, b, proximal = ProximalL1, x0, stepsize = stepsize, maxit = 100, abstol=1e-12, adaptive = True)
        sol, _ = fista.solve()

    """  
    def __init__(self, A, b, proximal, x0, maxit=100, stepsize=1e0, abstol=1e-14, adaptive = True):
        
        self.A = A
        self.b = b
        self.x0 = x0
        self.proximal = proximal
        self.maxit = int(maxit)
        self.stepsize = stepsize
        self.abstol = abstol
        self.adaptive = adaptive

    @property
    def _explicitA(self):
        return not callable(self.A)
            
    def solve(self):
        # initial state
        x = self.x0.copy()
        stepsize = self.stepsize
        
        k = 0
        
        while True:
            x_old = x.copy()
            k += 1
        
            if self._explicitA:
                grad = self.A.T@(self.A @ x_old - self.b)
            else:
                grad = self.A(self.A(x_old, 1) - self.b, 2)
                
            x_new = self.proximal(x_old-stepsize*grad, stepsize)
                        
            if LA.norm(x_new-x_old) <= self.abstol or (k >= self.maxit):
                return x_new, k
            
            if self.adaptive:
                x_new = x_new + ((k-1)/(k+2))*(x_new - x_old)
              
            x = x_new.copy()

class ADMM(object):
    """Alternating Direction Method of Multipliers for solving regularized linear least squares problems of the form:
    Minimize ||Ax-b||^2 + sum_i f_i(L_i x),
    where the sum ranges from 1 to an arbitrary n. See definition of the parameter `penalty_terms` below for more details about f_i and L_i

    Reference:
    [1] Boyd et al. "Distributed optimization and statistical learning via the alternating direction method of multipliers."Foundations and Trends® in Machine learning, 2011.

    
    Parameters
    ----------
    A : ndarray or callable
        Represents a matrix or a function that performs matrix-vector multiplications. 
        When A is a callable, it accepts arguments (x, flag) where:
        - flag=1 indicates multiplication of A with vector x, that is A @ x.
        - flag=2 indicates multiplication of the transpose of A with vector x, that is  A.T @ x.
    b : ndarray.
    penalty_terms : List of tuples (callable proximal operator of f_i, linear operator L_i)
        Each callable proximal operator of f_i accepts two arguments (x, p) and should return the minimizer of p/2||x-z||^2 + f(x) over z for some f.
    x0 : ndarray. Initial guess.
    penalty_parameter : Trade-off between linear least squares and regularization term in the solver iterates. Denoted as "rho" in [1].
    maxit : The maximum number of iterations.
    adaptive : Whether to adaptively update the penalty_parameter each iteration such that the primal and dual residual norms are of the same order of magnitude. Based on [1], Subsection 3.4.1
    
    Example
    -----------
    .. code-block:: python
    
        from cuqi.solver import ADMM, ProximalL1, ProjectNonnegative
        import numpy as np

        rng = np.random.default_rng()

        m, n, k = 10, 5, 4
        A = rng.standard_normal((m, n))
        b = rng.standard_normal(m)
        L = rng.standard_normal((k, n))

        x0 = np.zeros(n)
        admm = ADMM(A, b, x0, penalty_terms = [(ProximalL1, L), (lambda z, _ : ProjectNonnegative(z), np.eye(n))], tradeoff = 10)
        sol, _ = admm.solve()

    """  

    def __init__(self, A, b, penalty_terms, x0, penalty_parameter = 10, maxit = 100, inner_max_it = 10, adaptive = True):

        self.A = A
        self.b = b
        self.x_cur = x0

        dual_len = [penalty[1].shape[0] for penalty in penalty_terms]
        self.z_cur = [np.zeros(l) for l in dual_len]
        self.u_cur = [np.zeros(l) for l in dual_len]
        self.n = penalty_terms[0][1].shape[1]
        
        self.rho = penalty_parameter
        self.maxit = maxit
        self.inner_max_it = inner_max_it
        self.adaptive = adaptive

        self.penalty_terms = penalty_terms
       
        self.p = len(self.penalty_terms)
        self._big_matrix = None
        self._big_vector = None

    def solve(self):
        """
        Solves the regularized linear least squares problem using ADMM in scaled form. Based on [1], Subsection 3.1.1
        """
        z_new = self.p*[0]
        u_new = self.p*[0]

        # Iterating
        for i in range(self.maxit):
            self._iteration_pre_processing()

            # Main update (Least Squares)
            solver = CGLS(self._big_matrix, self._big_vector, self.x_cur, self.inner_max_it)
            x_new, _ = solver.solve()
        
            # Regularization update
            for j, penalty in enumerate(self.penalty_terms):
                z_new[j] = penalty[0](penalty[1]@x_new + self.u_cur[j], 1.0/self.rho)
                
            res_primal = 0.0
            # Dual update
            for j, penalty in enumerate(self.penalty_terms):
                r_partial = penalty[1]@x_new - z_new[j]
                res_primal += LA.norm(r_partial)**2

                u_new[j] = self.u_cur[j] + r_partial
            
            res_dual = 0.0
            for j, penalty in enumerate(self.penalty_terms):
                res_dual += LA.norm(penalty[1].T@(z_new[j] - self.z_cur[j]))**2

            # Adaptive approach based on [1], Subsection 3.4.1
            if self.adaptive:
                if res_dual > 1e2*res_primal:
                    self.rho *= 0.5 # More regularization
                elif res_primal > 1e2*res_dual:
                    self.rho *= 2.0 # More data fidelity

            self.x_cur, self.z_cur, self.u_cur = x_new, z_new.copy(), u_new
            
        return self.x_cur, i
    
    def _iteration_pre_processing(self):
            """ Preprocessing
            Every iteration of ADMM requires solving a linear least squares system of the form
                minimize 1/(rho) \|Ax-b\|_2^2 + sum_{i=1}^{p} \|penalty[1]x - (y - u)\|_2^2
            To solve this, all linear least squares terms are combined into a single big term
            with matrix big_matrix and data big_vector.

            The matrix only needs to be updated when rho changes, i.e., when the adaptive option is used.
            The data vector needs to be updated every iteration.
            """

            self._big_vector = np.hstack([np.sqrt(1/self.rho)*self.b] + [self.z_cur[i] - self.u_cur[i] for i in range(self.p)])

            # Check whether matrix needs to be updated
            if self._big_matrix is not None and not self.adaptive:
                return

            # Update big_matrix
            if callable(self.A):
                def matrix_eval(x, flag):
                    if flag == 1:
                        out1 = np.sqrt(1/self.rho)*self.A(x, 1)
                        out2 = [penalty[1]@x for penalty in self.penalty_terms]
                        out  = np.hstack([out1] + out2)
                    elif flag == 2:
                        idx_start = len(x)
                        idx_end = len(x)
                        out1 = np.zeros(self.n)
                        for _, t in reversed(self.penalty_terms):
                            idx_start -= t.shape[0]
                            out1 += t.T@x[idx_start:idx_end]
                            idx_end = idx_start
                        out2 = np.sqrt(1/self.rho)*self.A(x[:idx_end], 2)
                        out  = out1 + out2     
                    return out
                self._big_matrix = matrix_eval
            else:
                self._big_matrix = np.vstack([np.sqrt(1/self.rho)*self.A] + [penalty[1] for penalty in self.penalty_terms])




def ProjectNonnegative(x):
    """(Euclidean) projection onto the nonnegative orthant.
    
    Parameters
    ----------
    x : array_like.
    """  
    return np.maximum(x, 0)

def ProjectBox(x, lower = None, upper = None):
    """(Euclidean) projection onto a box.
    
    Parameters
    ----------
    x : array_like.
    lower : array_like. Lower bound of box. Zero if None.
    upper : array_like. Upper bound of box. One if None.
    """  
    if lower is None:
        lower = np.zeros_like(x)
    
    if upper is None:
        upper = np.ones_like(x)
    
    return np.minimum(np.maximum(x, lower), upper)

def ProjectHalfspace(x, a, b):
    """(Euclidean) projection onto the halfspace defined {z|<a,z> <= b}.
    
    Parameters
    ----------
    x : array_like.
    a : array_like.
    b : array_like.
    """  

    ax_b = np.inner(a,x) - b
    if ax_b <= 0:
        return x
    else:
        return x - (ax_b/np.inner(a,a))*a

def ProximalL1(x, gamma):
    """(Euclidean) proximal operator of the \|x\|_1 norm.
    Also known as the shrinkage or soft thresholding operator.
    
    Parameters
    ----------
    x : array_like.
    gamma : scale parameter.
    """
    return np.multiply(np.sign(x), np.maximum(np.abs(x)-gamma, 0))



import scipy as sp
import numpy as np
import cuqi
from cuqi.solver import CGLS
from cuqi.experimental.mcmc import Sampler

class UGLA(Sampler):
    """ Unadjusted (Gaussian) Laplace Approximation sampler
    
    Samples an approximate posterior where the prior is approximated
    by a Gaussian distribution. The likelihood must be Gaussian.

    Currently only works for LMRF priors.

    The inner solver is Conjugate Gradient Least Squares (CGLS) solver.

    For more details see: Uribe, Felipe, et al. A hybrid Gibbs sampler for edge-preserving 
    tomographic reconstruction with uncertain view angles. SIAM/ASA Journal on UQ,
    https://doi.org/10.1137/21M1412268 (2022).

    Parameters
    ----------
    target : `cuqi.distribution.Posterior`
        The target posterior distribution to sample.

    initial_point : ndarray, *Optional*
        Initial parameters.
        If not provided, it defaults to zeros.

    maxit : int
        Maximum number of inner iterations for solver when generating one sample.
        If not provided, it defaults to 50.

    tol : float
        Tolerance for inner solver.
        The inner solvers will stop before maxit if convergence check reaches tol.
        If not provided, it defaults to 1e-4.

    beta : float
        Smoothing parameter for the Gaussian approximation of the Laplace distribution.
        A small value in the range of 1e-7 to 1e-3 is recommended, though values out of this 
        range might give better results in some cases. Generally, a larger beta value makes 
        sampling easier but results in a worse approximation. See details in Section 3.3 of the paper.
        If not provided, it defaults to 1e-5.

    callback : callable, optional
        A function that will be called after each sampling step. It can be useful for monitoring the sampler during sampling.
        The function should take three arguments: the sampler object, the index of the current sampling step, the total number of requested samples. The last two arguments are integers. An example of the callback function signature is: `callback(sampler, sample_index, num_of_samples)`.
    """
    def __init__(self, target=None, initial_point=None, maxit=50, tol=1e-4, beta=1e-5, **kwargs):

        super().__init__(target=target, initial_point=initial_point, **kwargs)

        # Parameters
        self.maxit = maxit
        self.tol = tol
        self.beta = beta
    
    def _initialize(self):
        self._precompute()

    @property
    def prior(self):
        return self.target.prior

    @property
    def likelihood(self):
        return self.target.likelihood

    @property
    def model(self):
        return self.target.model     
    
    @property
    def _data(self):
        return self.target.data - self.target.model._shift

    def _precompute(self):

        D = self.prior._diff_op
        n = D.shape[0]

        # Gaussian approximation of LMRF prior as function of x_k
        def Lk_fun(x_k):
            dd =  1/np.sqrt((D @ x_k)**2 + self.beta*np.ones(n))
            W = sp.sparse.diags(dd)
            return W.sqrt() @ D
        self.Lk_fun = Lk_fun

        self._m = len(self._data)
        self._L1 = self.likelihood.distribution.sqrtprec

        # If prior location is scalar, repeat it to match dimensions
        if len(self.prior.location) == 1:
            self._priorloc = np.repeat(self.prior.location, self.dim)
        else:
            self._priorloc = self.prior.location

        # Initial Laplace approx
        self._L2 = Lk_fun(self.initial_point)
        self._L2mu = self._L2@self._priorloc
        self._b_tild = np.hstack([self._L1@self._data, self._L2mu]) 
        
        # Least squares form
        def M(x, flag):
            if flag == 1:
                out1 = self._L1 @ self.model._forward_func_no_shift(x) # Use forward function which excludes shift
                out2 = np.sqrt(1/self.prior.scale)*(self._L2 @ x)
                out  = np.hstack([out1, out2])
            elif flag == 2:
                idx = int(self._m)
                out1 = self.model._adjoint_func_no_shift(self._L1.T@x[:idx])
                out2 = np.sqrt(1/self.prior.scale)*(self._L2.T @ x[idx:])
                out  = out1 + out2                
            return out
        self.M = M

    def step(self):
        # Update Laplace approximation
        self._L2 = self.Lk_fun(self.current_point)
        self._L2mu = self._L2@self._priorloc
        self._b_tild = np.hstack([self._L1@self._data, self._L2mu]) 
    
        # Sample from approximate posterior
        e = np.random.randn(len(self._b_tild))
        y = self._b_tild + e # Perturb data
        sim = CGLS(self.M, y, self.current_point, self.maxit, self.tol)                     
        self.current_point, _ = sim.solve()
        acc = 1
        return acc

    def tune(self, skip_len, update_count):
        pass
    
    def validate_target(self):
        # Check target type
        if not isinstance(self.target, cuqi.distribution.Posterior):
            raise ValueError(f"To initialize an object of type {self.__class__}, 'target' need to be of type 'cuqi.distribution.Posterior'.")       

        # Check Affine model
        if not isinstance(self.likelihood.model, cuqi.model.AffineModel):
            raise TypeError("Model needs to be affine or linear")

        # Check Gaussian likelihood
        if not hasattr(self.likelihood.distribution, "sqrtprec"):
            raise TypeError("Distribution in Likelihood must contain a sqrtprec attribute")

        # Check that prior is LMRF
        if not isinstance(self.prior, cuqi.distribution.LMRF):
            raise ValueError('Unadjusted Gaussian Laplace approximation (UGLA) requires LMRF prior')
        
    def _get_default_initial_point(self, dim):
        """ Get the default initial point for the sampler. Defaults to an array of zeros. """
        return np.zeros(dim)



import cuqi
import pytest
import numpy as np

geom = {"geometry": 1}

#TODO. Make tests for all distributions going through their input variables
@pytest.mark.parametrize("dist, expected",[
    (cuqi.distribution.Gaussian(**geom), ["mean", "cov"]),
    (cuqi.distribution.Gaussian(mean=1, **geom), ["cov"]),
    (cuqi.distribution.Gaussian(cov=1, **geom), ["mean"]),
    (cuqi.distribution.Gaussian(mean=lambda m: m, cov=lambda c:c, **geom), ["m", "c"]),
    (cuqi.distribution.Gaussian(mean=1, cov=lambda c:c, **geom), ["c"]),
    (cuqi.distribution.Gaussian(mean=lambda m:m, cov=1, **geom), ["m"]),
])
def test_conditioning_variables(dist, expected):
    assert dist.get_conditioning_variables() == expected

def test_conditioning_through_likelihood(): #TODO. Add more dists to test here!
    """ This checks the flow of likelihood evaluation. In particular it checks:
        1) that the mean @property setter is invoked after model evaluation through function/model.
        2) that the cov @property setter is invoked after setting cov directly (here cov getter returned None).
    """
    model = cuqi.model.Model(lambda x:x, 1, 1) #Simple 1 par model
    dist = cuqi.distribution.Gaussian(mean=model)
    likelihood = dist.to_likelihood(5)
    likelihood.logd(10, 5)
    assert(likelihood.get_parameter_names() == ['cov', 'x'])

def test_conditioning_wrong_keyword():
    """ This checks if we raise error if keyword is not a conditioning variable. """
    x = cuqi.distribution.Gaussian(mean=1)
    with pytest.raises(ValueError):
        x(name="hey") #Should expect value-error since name not conditioning variable.

def test_conditioning_arg_as_mutable_var():
    """ This checks if we raise error if 2 args are given for a distribution that has no conditioning variables. """
    x = cuqi.distribution.Gaussian(mean=1, cov=1)
    with pytest.raises(ValueError):
        x(5, 3) #Should expect value-error since no cond vars

def test_conditioning_on_main_parameter():
    """ This checks if we can condition on the main parameter in various ways. """
    x = cuqi.distribution.Gaussian(geometry=1)

    # With keywords (name also automatically inferred)
    assert isinstance(x(mean=1, x=5), cuqi.likelihood.Likelihood)
    assert isinstance(x(mean=1, cov=1, x=5), cuqi.density.EvaluatedDensity)
    
    # Now using positional arguments
    assert isinstance(x(1, 1, 5), cuqi.density.EvaluatedDensity)

def test_conditioning_kwarg_as_mutable_var():
    """ This checks if we allow kwargs for a distribution that has no conditioning variables. """
    x = cuqi.distribution.Gaussian(mean=1, cov=1)
    with pytest.raises(ValueError):
        x = x(cov=2) #This should raise error since no cond vars

def test_conditioning_both_args_kwargs():
    """ This tests that we throw error if we accidentally provide arg and kwarg for same variable. """
    x = cuqi.distribution.Gaussian(mean=1)
    with pytest.raises(ValueError):
        x(1, cov=1) #Should expect value-error since cov is given as arg and kwarg.

def test_conditioning_multiple_args():
    """ This tests if conditional variables are printed correctly if they appear in multiple mutable variables"""
    dist = cuqi.distribution.Gaussian(mean=lambda x,y: x+y, cov=lambda y,z: y+z, geometry=1)
    assert dist.get_conditioning_variables() == ["x","y","z"]

def test_conditioning_partial_function():
    """ This tests the partial functions we define if only part of a callable is given. """
    dist = cuqi.distribution.Gaussian(mean=lambda x,y: x+y, cov=lambda y,z: y+z, geometry=1)
    dist2 = dist(x=1, y=2)
    assert dist2.get_conditioning_variables() == ["z"]
    assert dist2.mean == 3
    assert cuqi.utilities.get_non_default_args(dist2.cov) == ["z"]

def test_conditioning_keeps_name():
    """ This tests if the name of the distribution is kept when conditioning. """
    y = cuqi.distribution.Gaussian(lambda x:x, geometry=1)
    
    assert y(x=1, cov=1).name == y.name
    assert y(x=1)(cov=1).name == y.name
    assert y(x=1)(cov=1)().name == y.name

    assert y(x=1).name == y.name
    assert y(cov=1).name == y.name
    assert y().name == y.name

def test_conditioning_class_flow():
    """ This tests the class flow of conditioning a conditional distribution in various ways """

    # Initial conditional distribution on parameter x and cov.
    y = cuqi.distribution.Gaussian(lambda x:x, geometry=1)

    # Conditioning on x is still conditional on cov
    assert y(x=1).is_cond

    # Conditioning on y (name of distribution) should return a likelihood
    assert isinstance(y(y=1), cuqi.likelihood.Likelihood)

    # Conditioning on x and cov be a regular distribution
    assert isinstance(y(x=1, cov=1), cuqi.distribution.Distribution)

    # Conditioning on all unspecified variables should return constant density
    assert isinstance(y(x=1, y=1, cov=1), cuqi.density.EvaluatedDensity)

    # Conditioning on all parameters in various ways should also return constant density
    # (in between they might change to Likelihood)
    assert isinstance(y(x=1)(y=1)(cov=1), cuqi.density.EvaluatedDensity)
    assert isinstance(y(x=1)(cov=1)(y=1), cuqi.density.EvaluatedDensity)
    assert isinstance(y(y=1)(x=1)(cov=1), cuqi.density.EvaluatedDensity)
    assert isinstance(y(y=1)(cov=1)(x=1), cuqi.density.EvaluatedDensity)
    assert isinstance(y(cov=1)(x=1)(y=1), cuqi.density.EvaluatedDensity)
    assert isinstance(y(cov=1)(y=1)(x=1), cuqi.density.EvaluatedDensity)

def test_logp_conditional():
    """ This tests logp evaluation for conditional distributions """
    # Base example logp value
    true_val = cuqi.distribution.Gaussian(3, 7).logd(13)

    # Distribution with no specified parameters
    x = cuqi.distribution.Gaussian(cov=lambda s:s, geometry=1)

    # Test logp evaluates correctly in various cases
    assert x.logd(mean=3, s=7, x=13) == true_val
    assert x(x=13).logd(mean=3, s=7) == true_val
    assert x(x=13, mean=3).logd(s=7) == true_val
    assert x(x=13, mean=3, s=7).logd() == true_val
    assert x(mean=3).logd(s=7, x=13) == true_val
    assert x(mean=3, s=7).logd(x=13) == true_val
    assert x(mean=3, x=13).logd(s=7) == true_val

def test_logd_err_handling():
    """ This tests if logp correctly identifies errors in the input """
    x = cuqi.distribution.Gaussian(cov=lambda s:s, geometry=1)

    # Test that we raise error if we don't provide all parameters
    with pytest.raises(ValueError, match=r"To evaluate the log density all conditioning variables and main"):
        x.logd(x=3)

    # Test that we raise error if we provide parameters that are not specified
    with pytest.raises(ValueError, match=r"do not match keyword arguments"):
        x.logd(mean=3, s=7, x=13, y=1)

def test_logd_err_handling_single_cond_var():
    """ This tests if logp correctly identifies errors in the input """
    x = cuqi.distribution.Gaussian(0, cov=lambda s:s)

    # Test that we raise error if we don't provide all parameters
    with pytest.raises(ValueError, match=r"To evaluate the log density all conditioning variables and main"):
        x.logd(3) # Should expect error since we have not specified s

    # Too many arguments
    with pytest.raises(ValueError, match=r"Unable to parse"):
        x.logd(1, 3, 5)

    # Should not expect error since we have specified s (2 arguments)
    x.logd(1, 3)
    x.logd(s=1, x=1) 

def test_sample_conditional_err_handling():
    """ Test if conditional distributions correctly identify errors when sampling """
    x = cuqi.distribution.Gaussian(0, lambda s:s)

    # Test that we raise error if we don't provide all parameters
    with pytest.raises(ValueError, match=r"Missing conditioning variables:"):
        x.sample()


def test_cond_positional_and_kwargs():
    """ Test conditioning for both positional and kwargs """
    x = cuqi.distribution.Gaussian(cov=lambda s:s, geometry=1)

    logd = x(mean=3, s=7).logd(13)

    # Conditioning full positional
    assert x(3, 7, 13).value == logd
    assert x(3, 7)(13).value == logd
    assert x(3)(7, 13).value == logd
    assert x(3)(7)(13).value == logd

    # Conditioning full kwargs
    assert x(mean=3, s=7, x=13).value == logd
    assert x(mean=3, s=7)(x=13).value == logd
    assert x(mean=3)(s=7, x=13).value == logd
    assert x(mean=3)(s=7)(x=13).value == logd

    # Conditioning partial positional
    assert x(3, s=7, x=13).value == logd
    assert x(3, 7, x=13).value == logd
    assert x(3, s=7)(13).value == logd
    assert x(mean=3)(7, x=13).value == logd
    assert x(mean=3)(7)(x=13).value == logd
    assert x(mean=3)(7)(13).value == logd

def test_logd_positional_and_kwargs():
    """ Test logd for both positional and kwargs """
    x = cuqi.distribution.Gaussian(cov=lambda s:s, geometry=1)

    logd = x(mean=3, s=7).logd(13)

    # logd full kwargs
    assert x.logd(mean=3, s=7, x=13) == logd

    # logd full positional
    assert x.logd(3, 7, 13) == logd

    # logd partial positional
    assert x.logd(3, s=7, x=13) == logd
    assert x.logd(3, 7, x=13) == logd

def test_dim_geometry_compatibility():
    """ Test the compatibility of dim and geometry attributes """
    dist = cuqi.distribution.Gaussian(geometry=(2,2))
    assert dist.dim == dist.geometry.par_dim

def test_geometry_inference_from_dim():
    """ Test that the geometry attribute is correctly inferred from dim """
    dist = cuqi.distribution.Gaussian(geometry=5)
    assert dist.dim == 5
    assert dist.geometry.par_dim == 5

def test_geometry_inference_from_variables():
    """ Test that the geometry attribute is correctly inferred from variables """
    dist = cuqi.distribution.Gaussian(mean=np.ones(5))
    assert dist.dim == 5
    assert dist.geometry.par_dim == 5

def test_geometry_setting():
    """ Test that geometry attribute can be set and is correctly reflected in dim """
    dist = cuqi.distribution.Gaussian(geometry=1)
    dist.geometry = cuqi.geometry.Continuous1D(np.ones(5))
    assert dist.dim == 5
    assert dist.geometry.par_dim == 5

def test_dim_geometry_conflict():
    """ Test that an error is raised when both dim and geometry are specified """
    with pytest.raises(TypeError, match=r"Inconsistent distribution geometry"):
        dist = cuqi.distribution.Gaussian(np.zeros(2), geometry=cuqi.geometry.Continuous1D(np.ones(5)))
        


